#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2022 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License

import os
import tempfile
import time
from typing import Any

import numpy as np
import tensorflow as tf

from realbook.callbacks.utilization import CpuUtilizationCallback, MemoryUtilizationCallback
from realbook.layers.signal import Spectrogram


class SleeperCallback(tf.keras.callbacks.Callback):
    def __init__(self, duration: float) -> None:
        self.duration = duration

    def on_train_begin(self, logs: Any) -> None:
        time.sleep(self.duration)


TEST_AUDIO = np.linspace(0, 1, num=22050 * 10)


def test_cpu_memory_utilization() -> None:
    fake_data = tf.data.Dataset.zip(
        (
            tf.data.Dataset.from_tensor_slices([TEST_AUDIO]),
            tf.data.Dataset.from_tensor_slices([1]),
        )
    ).batch(1)

    model = tf.keras.Sequential(
        [
            tf.keras.Input(shape=(None,)),
            Spectrogram(),
            # Add a final "channel" dimension after the spectrogram
            tf.keras.layers.Lambda(lambda x: tf.expand_dims(x, -1)),
            tf.keras.layers.Dense(1),
        ]
    )
    model.compile(loss="binary_crossentropy")

    with tempfile.TemporaryDirectory() as tmpdir:
        writer = tf.summary.create_file_writer(tmpdir)
        cpu_callback = CpuUtilizationCallback(
            writer,
            1,
        )

        memory_callback = MemoryUtilizationCallback(
            writer,
            1,
        )

        model.fit(fake_data, callbacks=[cpu_callback, memory_callback, SleeperCallback(5)])

        # A check to make sure we're actually logging something
        # It'd be nice to also make assertions on the cpu and memory percentage, but
        # that's unreliable as it varies from machine to machine.
        cpu_count = 0
        memory_count = 0
        for e in tf.compat.v1.train.summary_iterator(os.path.join(tmpdir, os.listdir(tmpdir)[0])):
            for v in e.summary.value:
                cpu_count += CpuUtilizationCallback.RESOURCE_NAME in v.tag
                memory_count += MemoryUtilizationCallback.RESOURCE_NAME in v.tag
        assert cpu_count > 3 and cpu_count < 10
        assert memory_count > 3 and memory_count < 10



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2022 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Any, Iterable, Tuple, Union, Callable
import numpy as np
import io
import tensorflow as tf

try:
    import matplotlib.pyplot as plt
    import librosa.display
except ImportError as e:
    raise ImportError(str(e) + " (did you 'pip3 install realbook[tensorboard]'?)")


def plot_to_image(figure: plt.figure) -> tf.Tensor:
    """Converts the matplotlib plot specified by 'figure' to a PNG image and
    returns it. The supplied figure is closed and inaccessible after this call."""
    buf = io.BytesIO()
    plt.savefig(buf, format="png")
    plt.close(figure)
    buf.seek(0)
    return tf.expand_dims(tf.image.decode_png(buf.getvalue(), channels=4), 0)


class SpectrogramVisualizationCallback(tf.keras.callbacks.Callback):
    """
    If your model trains on image-like (i.e.: Spectrogram) data, this callback will:
     - identify which layers of your model are the "input" layers (by looking at their trainability)
     - interpret the output of the last non-trainable layer as a spectrogram
     - plot that spectrogram to TensorBoard on training start

    If your model has multiple branching paths in its frontend (i.e.: it produces two spectrograms)
    this callback will only pick the first. Subclass or copy this.

    Parameters:
     - tensorboard_writer: a TensorBoard SummaryWriter to use
     - example_batches: an iterator providing pairs of (batch, label) training data to visualize.
                        can be a tf.data.Dataset.
     - convert_to_dB: a callable to use to convert the post-spectrogram data to dB for Librosa
                      visualization. By default, this is librosa.amplitude_to_db - you may want to
                      use librosa.power_to_db instead if your spectrogram is a power spectrogram.
     - sample_rate: the sample rate of the input audio.
     - raise_on_error: if true, throw an exception if this callback causes an error.
                       by default, log exceptions but don't interrupt training.
     - name: the name of the summary data sent to TensorBoard.
     - Any remaining keyword arguments are passed through to librosa.display.specshow.
    """

    def __init__(
        self,
        tensorboard_writer: tf.summary.SummaryWriter,
        example_batches: Iterable[Tuple[tf.Tensor, tf.Tensor]],
        convert_to_dB: Union[bool, Callable[[tf.Tensor], tf.Tensor]] = True,
        sample_rate: int = 22050,
        # By default log exceptions but don't halt the training process.
        raise_on_error: bool = False,
        name: str = "Input to First Trainable Layer",
        **kwargs: Any,
    ):
        self.tensorboard_writer = tensorboard_writer
        self.example_batches = example_batches
        self.convert_to_dB = convert_to_dB
        self.sample_rate_hz = sample_rate
        self.raise_on_error = raise_on_error
        self.name = name

        self.specshow_arguments = kwargs
        if "x_axis" not in self.specshow_arguments:
            self.specshow_arguments["x_axis"] = "time"

    def on_train_begin(self, logs: Any = None) -> None:
        try:
            # Create a tempoary model using only the frontend of the model,
            # as defined by "the largest sequence of non-trainable layers at the start."
            non_trainable_input_layers = []
            for layer in self.model.layers:
                if len(layer.trainable_variables) + len(layer.trainable_weights) > 0:
                    break
                else:
                    non_trainable_input_layers.append(layer)

            if not non_trainable_input_layers:
                raise ValueError("No non-trainable input layers could be inferred for spectrogram visualization.")

            # Don't use tf.keras.models.Sequential here, as the input may not be traditional Layers.
            # (Yes, you'd think that self.model.layers returns all layers - but that doesn't seem to be the case.)
            input_to_image = tf.keras.models.Model(
                inputs=non_trainable_input_layers[0].input, outputs=non_trainable_input_layers[-1].output
            )

            with self.tensorboard_writer.as_default():
                # Pull n random batches from the dataset and send them to TensorBoard.
                for (data, _) in self.example_batches:
                    assert tf.rank(data) == 2, "Expected input data to be of rank 2, with shape (batch, audio)."
                    assert tf.shape(data)[0] < tf.shape(data)[1], (
                        "Expected input data to be of rank 2, with shape (batch, audio), but got shape"
                        f" {tf.shape(data)}."
                    )

                    spectrograms = input_to_image(data)
                    assert tf.rank(spectrograms) in (3, 4), (
                        "Expected non-trainable input layers to produce output of shape (batch, x, y) "
                        f"or (batch, x, y, 1), but got {tf.shape(spectrograms)}"
                    )
                    if tf.rank(spectrograms) == 4:
                        assert tf.shape(spectrograms)[-1] == 1, (
                            "Expected non-trainable input layers to produce output with one channel, but shape is"
                            f" {tf.shape(spectrograms)}"
                        )
                        # Ignore the single channel dimension, if it exists.
                        spectrograms = spectrograms[:, :, :, 0]

                    # We can infer the hop length, as we know the input audio length
                    # and sample rate used in the spectrogram
                    length_in_samples = data.shape[-1]
                    length_in_frames = spectrograms.shape[-2]
                    hop_length = int(tf.math.ceil(length_in_samples / length_in_frames))

                    figs = []
                    for spectrogram in spectrograms:
                        plt.clf()
                        fig, ax = plt.subplots()
                        spectrogram = np.abs(spectrogram).T

                        if self.convert_to_dB is True:
                            spectrogram = librosa.amplitude_to_db(spectrogram, ref=np.max)
                        elif callable(self.convert_to_dB):
                            spectrogram = self.convert_to_dB(spectrogram)

                        librosa.display.specshow(
                            spectrogram,
                            sr=self.sample_rate_hz,
                            hop_length=hop_length,
                            ax=ax,
                            **self.specshow_arguments,
                        )

                        figs.append(plot_to_image(fig))
                    tf.summary.image(
                        self.name,
                        np.concatenate(figs),
                        step=0,  # We only output this once, so epoch doesn't matter.
                        max_outputs=1000000,
                    )
                    plt.clf()
            self.tensorboard_writer.flush()
        except Exception as e:
            if self.raise_on_error:
                raise
            logging.error(f"{self.__class__.__name__} failed: ", exc_info=e)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2022 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import tensorflow as tf

from realbook.layers import math


def test_normalized_log_basic() -> None:
    norm_log = tf.keras.Sequential([tf.keras.layers.InputLayer((1, 10, 10)), math.NormalizedLog()])
    norm_log.compile()
    x = tf.random.normal([1, 1, 10, 10], mean=5, stddev=10)
    norm_log = norm_log(x).numpy()
    assert np.min(norm_log) >= 0
    assert np.max(norm_log) <= 1


def test_normalized_log_already_scaled() -> None:
    x = tf.random.normal([1, 10, 10], mean=0.1, stddev=0.1)
    norm_log = math.NormalizedLog()(x).numpy()
    assert np.min(norm_log) >= 0
    assert np.max(norm_log) <= 1


def test_normalized_log_zero_centered() -> None:
    x = tf.random.normal([1, 10, 10], mean=0, stddev=1)
    norm_log = math.NormalizedLog()(x).numpy()
    assert np.min(norm_log) >= 0
    assert np.max(norm_log) <= 1


def test_normalized_log_negative() -> None:
    x = tf.random.normal([1, 10, 10], mean=-5, stddev=1)
    norm_log = math.NormalizedLog()(x).numpy()
    assert np.min(norm_log) >= 0
    assert np.max(norm_log) <= 1


def test_normalized_log_batch() -> None:
    """Test that this is applied channel wise and not
    across the whole batch
    """
    gram1 = 0.5 * tf.ones([1, 2, 2])
    norm1 = math.NormalizedLog()(gram1).numpy()
    gram2 = tf.ones([1, 2, 2])
    norm2 = math.NormalizedLog()(gram2).numpy()
    gram3 = 0.25 * tf.ones([1, 2, 2])
    norm3 = math.NormalizedLog()(gram3).numpy()
    gram4 = tf.zeros([1, 2, 2])
    norm4 = math.NormalizedLog()(gram4).numpy()
    gram_batch = tf.concat([gram1, gram2, gram3, gram4], axis=0)
    norm_batch = math.NormalizedLog()(gram_batch).numpy()

    assert np.allclose(norm_batch[0, :, :], norm1[0, :, :], rtol=0)
    assert np.allclose(norm_batch[1, :, :], norm2[0, :, :], rtol=0)
    assert np.allclose(norm_batch[2, :, :], norm3[0, :, :], rtol=0)
    assert np.allclose(norm_batch[3, :, :], norm4[0, :, :], rtol=0)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2022 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import faulthandler
import os
import signal
import sys
import tensorflow as tf
import time
import threading
from typing import Any, List, Optional


class HangDebugCallback(tf.keras.callbacks.Callback):
    """
    Sometimes, when training models, training jobs can hang for unknown reasons.
    This callback will dump stack traces to standard error if an epoch takes more
    than 2x the average time, and will exit the process if an epoch takes longer
    than 5x the average time.

    Useful when training on cloud-based training systems that don't allow for sending
    SIGINT/KeyboardInterrupt/Ctrl-C to a process.
    """

    CHECK_INTERVAL_SECONDS: float = 1.0

    def __init__(
        self,
        print_stacks_if_epoch_time_exceeds_multiple: float = 2,
        abort_if_epoch_time_exceeds_multiple: float = 5,
    ):
        """
        Create a new HangDebugCallback.

        Args:
            print_stacks_if_epoch_time_exceeds_multiple (float):
                if a single epoch takes this many times longer than the average,
                stack traces of all threads will be printed to the standard error stream.
            abort_if_epoch_time_exceeds_multiple (float):
                if a single epoch takes this many times longer than the average,
                this process will be exited with return code 6 (SIGABRT).
        """
        self.monitoring_thread = threading.Thread(target=self.monitoring_fn, name="hang_debug_thread", daemon=True)
        self.epoch_start_time: Optional[float] = None
        self.finished = False
        self.epoch_times: List[float] = []
        self.dumped_stacks_already = False
        self.print_stacks_if_epoch_time_exceeds_multiple = print_stacks_if_epoch_time_exceeds_multiple
        self.abort_if_epoch_time_exceeds_multiple = abort_if_epoch_time_exceeds_multiple

    def on_train_begin(self, logs: Any) -> None:
        """
        Runs at the beginning of training and starts the measurement thread.
        """
        self.monitoring_thread.start()

    def on_train_end(self, logs: Any) -> None:
        """
        Runs at the end of training. Stops the measurement thread and flushes any buffered
        measurements.
        """
        self.finished = True

    def on_epoch_begin(self, epoch: int, logs: Any) -> None:
        self.epoch_start_time = time.time()

    def on_epoch_end(self, epoch: int, logs: Any) -> None:
        if self.current_epoch_duration:
            self.epoch_times.append(self.current_epoch_duration)

    @property
    def current_epoch_duration(self) -> Optional[float]:
        if self.epoch_start_time is not None:
            return time.time() - self.epoch_start_time
        return None

    @property
    def mean_epoch_time(self) -> Optional[float]:
        if not self.epoch_times:
            return None
        return sum(self.epoch_times) / len(self.epoch_times)

    @property
    def should_dump_stacks(self) -> bool:
        if not self.current_epoch_duration or not self.mean_epoch_time:
            return False
        return self.current_epoch_duration > (self.mean_epoch_time * self.print_stacks_if_epoch_time_exceeds_multiple)

    @property
    def should_abort(self) -> bool:
        if not self.current_epoch_duration or not self.mean_epoch_time:
            return False
        return self.current_epoch_duration > (self.mean_epoch_time * self.abort_if_epoch_time_exceeds_multiple)

    def abort(self) -> None:
        # AI Platform will retry the job if we raise SIGABRT.
        # We're in a thread here, so we need to kill the whole
        # process with os._exit - otherwise we'll just kill the
        # current thread and not stop the process.
        os._exit(signal.SIGABRT)

    def monitoring_fn(self) -> None:
        while not self.finished:
            if self.should_dump_stacks and not self.dumped_stacks_already and self.mean_epoch_time:
                sys.stderr.write(
                    f"Mean epoch time is {self.mean_epoch_time:2.2f} "
                    "seconds, but current epoch is taking "
                    f"{self.current_epoch_duration:2.2f} seconds. "
                    "Printing stacks for all threads:\n"
                )
                sys.stderr.flush()
                faulthandler.dump_traceback()
                sys.stderr.flush()
                sys.stderr.write(
                    "If the current epoch does not finish within "
                    f"{5 * self.mean_epoch_time} seconds, this job "
                    "will exit.\n"
                )
                sys.stderr.flush()
                self.dumped_stacks_already = True
            if self.should_abort:
                sys.stderr.write(
                    f"Mean epoch time is {self.mean_epoch_time:2.2f} "
                    "seconds, but current epoch is taking "
                    f"{self.current_epoch_duration:2.2f} seconds. "
                    "Aborting job.\n"
                )
                sys.stderr.flush()
                self.abort()
            time.sleep(self.CHECK_INTERVAL_SECONDS)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2022 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module is comprised of PyTorch layers copied from NNAudio and ported to TensorFlow:
https://github.com/KinWaiCheuk/nnAudio

A lot of this code is of questionable quality.
It was copied as correctly as possible from the original PyTorch code,
but bugs in the original code may have come along for the ride.
"""

import warnings
import tensorflow as tf
import numpy as np
from typing import Any, Callable, Dict, Optional, Tuple, Union

import scipy.signal


def create_lowpass_filter(
    band_center: float = 0.5,
    kernel_length: int = 256,
    transition_bandwidth: float = 0.03,
    dtype: tf.dtypes.DType = tf.float32,
) -> np.ndarray:
    """
    Calculate the highest frequency we need to preserve and the lowest frequency we allow
    to pass through. Note that frequency is on a scale from 0 to 1 where 0 is 0 and 1 is
    the Nyquist frequency of the signal BEFORE downsampling.
    """

    passband_max = band_center / (1 + transition_bandwidth)
    stopband_min = band_center * (1 + transition_bandwidth)

    # We specify a list of key frequencies for which we will require
    # that the filter match a specific output gain.
    # From [0.0 to passband_max] is the frequency range we want to keep
    # untouched and [stopband_min, 1.0] is the range we want to remove
    key_frequencies = [0.0, passband_max, stopband_min, 1.0]

    # We specify a list of output gains to correspond to the key
    # frequencies listed above.
    # The first two gains are 1.0 because they correspond to the first
    # two key frequencies. the second two are 0.0 because they
    # correspond to the stopband frequencies
    gain_at_key_frequencies = [1.0, 1.0, 0.0, 0.0]

    # This command produces the filter kernel coefficients
    filter_kernel = scipy.signal.firwin2(kernel_length, key_frequencies, gain_at_key_frequencies)

    return tf.constant(filter_kernel, dtype=dtype)


def next_power_of_2(A: int) -> int:
    """A helper function to calculate the next nearest number to the power of 2."""
    return int(np.ceil(np.log2(A)))


def early_downsample(
    sr: int,
    hop_length: int,
    n_octaves: int,
    nyquist_hz: float,
    filter_cutoff_hz: float,
) -> Tuple[int, int, int]:
    """Return new sampling rate and hop length after early downsampling"""
    downsample_count = early_downsample_count(nyquist_hz, filter_cutoff_hz, hop_length, n_octaves)
    downsample_factor = 2 ** (downsample_count)

    hop_length //= downsample_factor  # Getting new hop_length
    new_sr = sr / float(downsample_factor)  # Getting new sampling rate
    sr = int(new_sr)

    return sr, hop_length, downsample_factor


# The following two downsampling count functions are obtained from librosa CQT
# They are used to determine the number of pre resamplings if the starting and ending frequency
# are both in low frequency regions.
def early_downsample_count(nyquist_hz: float, filter_cutoff_hz: float, hop_length: int, n_octaves: int) -> int:
    """Compute the number of early downsampling operations"""

    downsample_count1 = max(0, int(np.ceil(np.log2(0.85 * nyquist_hz / filter_cutoff_hz)) - 1) - 1)
    num_twos = next_power_of_2(hop_length)
    downsample_count2 = max(0, num_twos - n_octaves + 1)

    return min(downsample_count1, downsample_count2)


def get_early_downsample_params(
    sr: int,
    hop_length: int,
    fmax_t: float,
    Q: float,
    n_octaves: int,
    dtype: tf.dtypes.DType,
) -> Tuple[int, int, int, Optional[np.array], bool]:
    """Compute downsampling parameters used for early downsampling"""

    window_bandwidth = 1.5  # for hann window
    filter_cutoff = fmax_t * (1 + 0.5 * window_bandwidth / Q)
    sr, hop_length, downsample_factor = early_downsample(sr, hop_length, n_octaves, sr // 2, filter_cutoff)
    if downsample_factor != 1:
        earlydownsample = True
        early_downsample_filter = create_lowpass_filter(
            band_center=1 / downsample_factor,
            kernel_length=256,
            transition_bandwidth=0.03,
            dtype=dtype,
        )
    else:
        early_downsample_filter = None
        earlydownsample = False

    return sr, hop_length, downsample_factor, early_downsample_filter, earlydownsample


def get_window_dispatch(window: Union[str, Tuple[str, int]], N: int, fftbins: bool = True) -> np.array:
    if isinstance(window, str):
        return scipy.signal.get_window(window, N, fftbins=fftbins)
    elif isinstance(window, tuple):
        if window[0] == "gaussian":
            assert window[1] >= 0
            sigma = np.floor(-N / 2 / np.sqrt(-2 * np.log(10 ** (-window[1] / 20))))
            return scipy.signal.get_window(("gaussian", sigma), N, fftbins=fftbins)
        else:
            Warning("Tuple windows may have undesired behaviour regarding Q factor")
    elif isinstance(window, float):
        Warning("You are using Kaiser window with beta factor " + str(window) + ". Correct behaviour not checked.")
    else:
        raise Exception("The function get_window from scipy only supports strings, tuples and floats.")


def create_cqt_kernels(
    Q: float,
    fs: int,
    fmin: float,
    n_bins: int = 84,
    bins_per_octave: int = 12,
    norm: Union[int, str] = 1,
    window: str = "hann",
    fmax: Optional[int] = None,
    topbin_check: bool = True,
) -> Tuple[np.array, int, np.array, np.array]:
    """
    Automatically create CQT kernels in time domain
    """

    fftLen = 2 ** next_power_of_2(np.ceil(Q * fs / fmin))

    if (fmax is not None) and (n_bins is None):
        n_bins = np.ceil(bins_per_octave * np.log2(fmax / fmin))  # Calculate the number of bins
        freqs = fmin * 2.0 ** (np.r_[0:n_bins] / float(bins_per_octave))

    elif (fmax is None) and (n_bins is not None):
        freqs = fmin * 2.0 ** (np.r_[0:n_bins] / float(bins_per_octave))

    else:
        warnings.warn("If fmax is given, n_bins will be ignored", SyntaxWarning)
        n_bins = np.ceil(bins_per_octave * np.log2(fmax / fmin))  # Calculate the number of bins
        freqs = fmin * 2.0 ** (np.r_[0:n_bins] / float(bins_per_octave))

    if np.max(freqs) > fs / 2 and topbin_check is True:
        raise ValueError(
            "The top bin {}Hz has exceeded the Nyquist frequency, please reduce the n_bins".format(np.max(freqs))
        )

    tempKernel = np.zeros((int(n_bins), int(fftLen)), dtype=np.complex64)

    lengths = np.ceil(Q * fs / freqs)
    for k in range(0, int(n_bins)):
        freq = freqs[k]
        _l = np.ceil(Q * fs / freq)

        # Centering the kernels, pad more zeros on RHS
        start = int(np.ceil(fftLen / 2.0 - _l / 2.0)) - int(_l % 2)

        sig = (
            get_window_dispatch(window, int(_l), fftbins=True)
            * np.exp(np.r_[-_l // 2 : _l // 2] * 1j * 2 * np.pi * freq / fs)
            / _l
        )

        if norm:  # Normalizing the filter # Trying to normalize like librosa
            tempKernel[k, start : start + int(_l)] = sig / np.linalg.norm(sig, norm)
        else:
            tempKernel[k, start : start + int(_l)] = sig

    return tempKernel, fftLen, lengths, freqs


def get_cqt_complex(
    x: tf.Tensor,
    cqt_kernels_real: tf.Tensor,
    cqt_kernels_imag: tf.Tensor,
    hop_length: int,
    padding: tf.keras.layers.Layer,
) -> tf.Tensor:
    """Multiplying the STFT result with the cqt_kernel, check out the 1992 CQT paper [1]
    for how to multiple the STFT result with the CQT kernel
    [2] Brown, Judith C.C. and Miller Puckette. “An efficient algorithm for the calculation of
    a constant Q transform.” (1992)."""

    # STFT, converting the audio input from time domain to frequency domain
    try:
        x = padding(x)  # When center is True, we need padding at the beginning and ending
    except Exception:
        warnings.warn(
            f"\ninput size = {x.shape}\tkernel size = {cqt_kernels_real.shape[-1]}\n"
            "padding with reflection mode might not be the best choice, try using constant padding",
            UserWarning,
        )
        x = tf.pad(x, (cqt_kernels_real.shape[-1] // 2, cqt_kernels_real.shape[-1] // 2))
    CQT_real = tf.transpose(
        tf.nn.conv1d(
            tf.transpose(x, [0, 2, 1]),
            tf.transpose(cqt_kernels_real, [2, 1, 0]),
            padding="VALID",
            stride=hop_length,
        ),
        [0, 2, 1],
    )
    CQT_imag = -tf.transpose(
        tf.nn.conv1d(
            tf.transpose(x, [0, 2, 1]),
            tf.transpose(cqt_kernels_imag, [2, 1, 0]),
            padding="VALID",
            stride=hop_length,
        ),
        [0, 2, 1],
    )

    return tf.stack((CQT_real, CQT_imag), axis=-1)


def downsampling_by_n(x: tf.Tensor, filter_kernel: tf.Tensor, n: int, match_torch_exactly: bool = True) -> tf.Tensor:
    """
    Downsample the given tensor using the given filter kernel.
    The input tensor is expected to have shape `(n_batches, channels, width)`,
    and the filter kernel is expected to have shape `(num_output_channels,)` (i.e.: 1D)

    If match_torch_exactly is passed, we manually pad the input rather than having TensorFlow do so with "SAME".
    The result is subtly different than Torch's output, but it is compatible with TensorFlow Lite (as of v2.4.1).
    """

    if match_torch_exactly:
        paddings = [
            [0, 0],
            [0, 0],
            [(filter_kernel.shape[-1] - 1) // 2, (filter_kernel.shape[-1] - 1) // 2],
        ]
        padded = tf.pad(x, paddings)

        # Store this tensor in the shape `(n_batches, width, channels)`
        padded_nwc = tf.transpose(padded, [0, 2, 1])
        result_nwc = tf.nn.conv1d(padded_nwc, filter_kernel[:, None, None], padding="VALID", stride=n)
    else:
        x_nwc = tf.transpose(x, [0, 2, 1])
        result_nwc = tf.nn.conv1d(x_nwc, filter_kernel[:, None, None], padding="SAME", stride=n)
    result_ncw = tf.transpose(result_nwc, [0, 2, 1])
    return result_ncw


class ReflectionPad1D(tf.keras.layers.Layer):
    """
    Replica of Torch's nn.ReflectionPad1D in TF.
    """

    def __init__(self, padding: int = 1, **kwargs: Any):
        self.padding = padding
        self.input_spec = [tf.keras.layers.InputSpec(ndim=3)]
        super(ReflectionPad1D, self).__init__(**kwargs)

    def compute_output_shape(self, s: Union[tf.TensorShape, Tuple[tf.Tensor, ...]]) -> tf.TensorShape:
        return (s[0], s[1], s[2] + 2 * self.padding)

    def call(self, x: tf.Tensor) -> tf.Tensor:
        return tf.pad(x, [[0, 0], [0, 0], [self.padding, self.padding]], "REFLECT")


class ConstantPad1D(tf.keras.layers.Layer):
    """
    Replica of Torch's nn.ConstantPad1D in TF.
    """

    def __init__(self, padding: int = 1, value: int = 0, **kwargs: Any):
        self.padding = padding
        self.value = value
        self.input_spec = [tf.keras.layers.InputSpec(ndim=3)]
        super(ConstantPad1D, self).__init__(**kwargs)

    def compute_output_shape(self, s: Union[tf.TensorShape, Tuple[tf.Tensor, ...]]) -> tf.TensorShape:
        return (s[0], s[1], s[2] + 2 * self.padding)

    def call(self, x: tf.Tensor) -> tf.Tensor:
        return tf.pad(x, [[0, 0], [0, 0], [self.padding, self.padding]], "CONSTANT", self.value)


def pad_center(data: np.ndarray, size: int, axis: int = -1, **kwargs: Any) -> np.ndarray:
    """Wrapper for np.pad to automatically center an array prior to padding.
    This is analogous to `str.center()`
    Examples
    --------
    >>> # Generate a vector
    >>> data = np.ones(5)
    >>> librosa.util.pad_center(data, 10, mode='constant')
    array([ 0.,  0.,  1.,  1.,  1.,  1.,  1.,  0.,  0.,  0.])
    >>> # Pad a matrix along its first dimension
    >>> data = np.ones((3, 5))
    >>> librosa.util.pad_center(data, 7, axis=0)
    array([[ 0.,  0.,  0.,  0.,  0.],
           [ 0.,  0.,  0.,  0.,  0.],
           [ 1.,  1.,  1.,  1.,  1.],
           [ 1.,  1.,  1.,  1.,  1.],
           [ 1.,  1.,  1.,  1.,  1.],
           [ 0.,  0.,  0.,  0.,  0.],
           [ 0.,  0.,  0.,  0.,  0.]])
    >>> # Or its second dimension
    >>> librosa.util.pad_center(data, 7, axis=1)
    array([[ 0.,  1.,  1.,  1.,  1.,  1.,  0.],
           [ 0.,  1.,  1.,  1.,  1.,  1.,  0.],
           [ 0.,  1.,  1.,  1.,  1.,  1.,  0.]])
    Parameters
    ----------
    data : np.ndarray
        Vector to be padded and centered
    size : int >= len(data) [scalar]
        Length to pad `data`
    axis : int
        Axis along which to pad and center the data
    kwargs : additional keyword arguments
      arguments passed to `np.pad()`
    Returns
    -------
    data_padded : np.ndarray
        `data` centered and padded to length `size` along the
        specified axis
    Raises
    ------
    ValueError
        If `size < data.shape[axis]`
    See Also
    --------
    numpy.pad
    """

    kwargs.setdefault("mode", "constant")

    n = data.shape[axis]

    lpad = int((size - n) // 2)

    lengths = [(0, 0)] * data.ndim
    lengths[axis] = (lpad, int(size - n - lpad))

    if lpad < 0:
        raise ValueError(("Target size ({:d}) must be at least input size ({:d})").format(size, n))

    return np.pad(data, lengths, **kwargs)


class CQT2010v2(tf.keras.layers.Layer):
    """This layer calculates the CQT of the input signal.
    Input signal should be in either of the following shapes.
    1. (len_audio)
    2. (num_audio, len_audio)
    3. (num_audio, 1, len_audio)
    The correct shape will be inferred autommatically if the input follows these 3 shapes.
    Most of the arguments follow the convention from librosa.

    This layer uses about 1MB of memory per second of input audio with its default arguments.

    This alogrithm uses the resampling method proposed in [1].
    Instead of convoluting the STFT results with a gigantic CQT kernel covering the full frequency
    spectrum, we make a small CQT kernel covering only the top octave. Then we keep downsampling the
    input audio by a factor of 2 to convoluting it with the small CQT kernel.
    Everytime the input audio is downsampled, the CQT relative to the downsampled input is equivalent
    to the next lower octave.
    The kernel creation process is still same as the 1992 algorithm. Therefore, we can reuse the
    code from the 1992 alogrithm [2]
    [1] Schörkhuber, Christian. “CONSTANT-Q TRANSFORM TOOLBOX FOR MUSIC PROCESSING.” (2010).
    [2] Brown, Judith C.C. and Miller Puckette. “An efficient algorithm for the calculation of a
    constant Q transform.” (1992).
    Early downsampling factor is to downsample the input audio to reduce the CQT kernel size.
    The result with and without early downsampling are more or less the same except in the very low
    frequency region where freq < 40Hz.
    Parameters
    ----------
    sr : int
        The sampling rate for the input audio. It is used to calucate the correct ``fmin`` and ``fmax``.
        Setting the correct sampling rate is very important for calculating the correct frequency.
    hop_length : int
        The hop (or stride) size. Default value is 512.
    fmin : float
        The frequency for the lowest CQT bin. Default is 32.70Hz, which coresponds to the note C0.
    fmax : float
        The frequency for the highest CQT bin. Default is ``None``, therefore the higest CQT bin is
        inferred from the ``n_bins`` and ``bins_per_octave``.  If ``fmax`` is not ``None``, then the
        argument ``n_bins`` will be ignored and ``n_bins`` will be calculated automatically.
        Default is ``None``
    n_bins : int
        The total numbers of CQT bins. Default is 84. Will be ignored if ``fmax`` is not ``None``.
    bins_per_octave : int
        Number of bins per octave. Default is 12.
    norm : bool
        Normalization for the CQT result.
    basis_norm : int
        Normalization for the CQT kernels. ``1`` means L1 normalization, and ``2`` means L2 normalization.
        Default is ``1``, which is same as the normalization used in librosa.
    window : str
        The windowing function for CQT. It uses ``scipy.signal.get_window``, please refer to
        scipy documentation for possible windowing functions. The default value is 'hann'
    pad_mode : str
        The padding method. Default value is 'reflect'.
    trainable : bool
        Determine if the CQT kernels are trainable or not. If ``True``, the gradients for CQT kernels
        will also be caluclated and the CQT kernels will be updated during model training.
        Default value is ``False``
    output_format : str
        Determine the return type.
        'Magnitude' will return the magnitude of the STFT result, shape = ``(num_samples, freq_bins, time_steps)``;
        'Complex' will return the STFT result in complex number, shape = ``(num_samples, freq_bins, time_steps, 2)``;
        'Phase' will return the phase of the STFT reuslt, shape = ``(num_samples, freq_bins,time_steps, 2)``.
        The complex number is stored as ``(real, imag)`` in the last axis. Default value is 'Magnitude'.
    verbose : bool
        If ``True``, it shows layer information. If ``False``, it suppresses all prints.
    device : str
        Choose which device to initialize this layer. Default value is 'cpu'.
    Returns
    -------
    spectrogram : tf.Tensor

    It returns a tensor of spectrograms.
    shape = ``(num_samples, freq_bins,time_steps)`` if ``output_format='Magnitude'``;
    shape = ``(num_samples, freq_bins,time_steps, 2)`` if ``output_format='Complex' or 'Phase'``;
    Examples
    --------
    >>> spec_layer = Spectrogram.CQT2010v2()
    >>> specs = spec_layer(x)
    """

    def __init__(
        self,
        sr: int = 22050,
        hop_length: int = 512,
        fmin: float = 32.70,
        fmax: Optional[float] = None,
        n_bins: int = 84,
        filter_scale: int = 1,
        bins_per_octave: int = 12,
        norm: bool = True,
        basis_norm: int = 1,
        window: str = "hann",
        pad_mode: str = "reflect",
        earlydownsample: bool = True,
        trainable: bool = False,
        output_format: str = "Magnitude",
        match_torch_exactly: bool = True,
    ):
        super().__init__()

        self.sample_rate = sr
        self.hop_length = hop_length
        self.fmin = fmin
        self.fmax = fmax
        self.n_bins = n_bins
        self.filter_scale = filter_scale
        self.bins_per_octave = bins_per_octave
        self.norm = norm  # Now norm is used to normalize the final CQT result by dividing n_fft
        self.basis_norm = basis_norm  # basis_norm is for normalizing basis
        self.window = window
        self.pad_mode = pad_mode
        # TODO: activate early downsampling later if possible
        self.earlydownsample = earlydownsample
        self.trainable = trainable
        self.output_format = output_format
        self.match_torch_exactly = match_torch_exactly
        self.normalization_type = "librosa"

    def get_config(self) -> Dict[str, Any]:
        config: Dict[str, Any] = super().get_config().copy()
        config.update(
            {
                "sample_rate": self.sample_rate,
                "hop_length": self.hop_length,
                "fmin": self.fmin,
                "fmax": self.fmax,
                "n_bins": self.n_bins,
                "filter_scale": self.filter_scale,
                "bins_per_octave": self.bins_per_octave,
                "norm": self.norm,
                "basis_norm": self.basis_norm,
                "window": self.window,
                "pad_mode": self.pad_mode,
                "output_format": self.output_format,
                "earlydownsample": self.earlydownsample,
                "trainable": self.trainable,
                "match_torch_exactly": self.match_torch_exactly,
            }
        )
        return config

    def build(self, input_shape: tf.TensorShape) -> None:
        # This will be used to calculate filter_cutoff and creating CQT kernels
        Q = float(self.filter_scale) / (2 ** (1 / self.bins_per_octave) - 1)

        self.lowpass_filter = create_lowpass_filter(band_center=0.5, kernel_length=256, transition_bandwidth=0.001)

        # Calculate num of filter requires for the kernel
        # n_octaves determines how many resampling requires for the CQT
        n_filters = min(self.bins_per_octave, self.n_bins)
        self.n_octaves = int(np.ceil(float(self.n_bins) / self.bins_per_octave))

        # Calculate the lowest frequency bin for the top octave kernel
        self.fmin_t = self.fmin * 2 ** (self.n_octaves - 1)
        remainder = self.n_bins % self.bins_per_octave

        if remainder == 0:
            # Calculate the top bin frequency
            fmax_t = self.fmin_t * 2 ** ((self.bins_per_octave - 1) / self.bins_per_octave)
        else:
            # Calculate the top bin frequency
            fmax_t = self.fmin_t * 2 ** ((remainder - 1) / self.bins_per_octave)

        self.fmin_t = fmax_t / 2 ** (1 - 1 / self.bins_per_octave)  # Adjusting the top minium bins
        if fmax_t > self.sample_rate / 2:
            raise ValueError(
                "The top bin {}Hz has exceeded the Nyquist frequency, please reduce the n_bins".format(fmax_t)
            )

        if self.earlydownsample is True:  # Do early downsampling if this argument is True
            (
                self.sample_rate,
                self.hop_length,
                self.downsample_factor,
                early_downsample_filter,
                self.earlydownsample,
            ) = get_early_downsample_params(self.sample_rate, self.hop_length, fmax_t, Q, self.n_octaves, self.dtype)

            self.early_downsample_filter = early_downsample_filter
        else:
            self.downsample_factor = 1

        # Preparing CQT kernels
        basis, self.n_fft, _, _ = create_cqt_kernels(
            Q,
            self.sample_rate,
            self.fmin_t,
            n_filters,
            self.bins_per_octave,
            norm=self.basis_norm,
            topbin_check=False,
        )

        # For the normalization in the end
        # The freqs returned by create_cqt_kernels cannot be used
        # Since that returns only the top octave bins
        # We need the information for all freq bin
        freqs = self.fmin * 2.0 ** (np.r_[0 : self.n_bins] / float(self.bins_per_octave))
        self.frequencies = freqs

        self.lengths = np.ceil(Q * self.sample_rate / freqs)

        self.basis = basis
        # NOTE(psobot): this is where the implementation here starts to differ from CQT2010.

        # These cqt_kernel is already in the frequency domain
        self.cqt_kernels_real = tf.expand_dims(basis.real.astype(self.dtype), 1)
        self.cqt_kernels_imag = tf.expand_dims(basis.imag.astype(self.dtype), 1)

        if self.trainable:
            self.cqt_kernels_real = tf.Variable(initial_value=self.cqt_kernels_real, trainable=True)
            self.cqt_kernels_imag = tf.Variable(initial_value=self.cqt_kernels_imag, trainable=True)

        # If center==True, the STFT window will be put in the middle, and paddings at the beginning
        # and ending are required.
        if self.pad_mode == "constant":
            self.padding: tf.keras.layers.Layer = ConstantPad1D(self.n_fft // 2, 0)
        elif self.pad_mode == "reflect":
            self.padding = ReflectionPad1D(self.n_fft // 2)

        rank = len(input_shape)
        if rank == 2:
            self.reshape_input: Callable[[tf.Tensor], tf.Tensor] = lambda x: x[:, None, :]
        elif rank == 1:
            self.reshape_input = lambda x: x[None, None, :]
        elif rank == 3:
            self.reshape_input = lambda x: x
        else:
            raise ValueError(f"Input shape must be rank <= 3, found shape {input_shape}")

    def call(self, x: tf.Tensor) -> tf.Tensor:
        x = self.reshape_input(x)

        if self.earlydownsample is True:
            x = downsampling_by_n(
                x,
                self.early_downsample_filter,
                self.downsample_factor,
                self.match_torch_exactly,
            )

        hop = self.hop_length

        # Getting the top octave CQT
        CQT = get_cqt_complex(x, self.cqt_kernels_real, self.cqt_kernels_imag, hop, self.padding)

        x_down = x  # Preparing a new variable for downsampling

        for i in range(self.n_octaves - 1):
            hop = hop // 2
            x_down = downsampling_by_n(x_down, self.lowpass_filter, 2, self.match_torch_exactly)
            CQT1 = get_cqt_complex(x_down, self.cqt_kernels_real, self.cqt_kernels_imag, hop, self.padding)
            CQT = tf.concat((CQT1, CQT), axis=1)

        CQT = CQT[:, -self.n_bins :, :]  # Removing unwanted bottom bins

        # Normalizing the output with the downsampling factor, 2**(self.n_octaves-1) is make it
        # same mag as 1992
        CQT = CQT * self.downsample_factor

        # Normalize again to get same result as librosa
        if self.normalization_type == "librosa":
            CQT *= tf.math.sqrt(tf.cast(self.lengths.reshape((-1, 1, 1)), self.dtype))
        elif self.normalization_type == "convolutional":
            pass
        elif self.normalization_type == "wrap":
            CQT *= 2
        else:
            raise ValueError("The normalization_type %r is not part of our current options." % self.normalization_type)

        # Transpose the output to match the output of the other spectrogram layers.
        if self.output_format.lower() == "magnitude":
            # Getting CQT Amplitude
            return tf.transpose(
                tf.math.sqrt(tf.math.reduce_sum(tf.math.pow(CQT, 2), axis=-1)),
                [0, 2, 1],
            )

        elif self.output_format.lower() == "complex":
            return CQT

        elif self.output_format.lower() == "phase":
            phase_real = tf.math.cos(tf.math.atan2(CQT[:, :, :, 1], CQT[:, :, :, 0]))
            phase_imag = tf.math.sin(tf.math.atan2(CQT[:, :, :, 1], CQT[:, :, :, 0]))
            return tf.stack((phase_real, phase_imag), axis=-1)


CQT = CQT2010v2



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2022 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tensorflow as tf
from google.protobuf.json_format import (
    MessageToDict as SerializeProtobufToDict,
    ParseDict as ParseDictToProtobuf,
)
from typing import Any, Dict, List, Optional, Tuple, Union
from tensorflow.python.framework.convert_to_constants import (
    convert_variables_to_constants_v2_as_graph,
)
from tensorflow.python.ops.op_selector import UnliftableError
from tensorflow.python.eager.wrap_function import WrappedFunction


def _load_concrete_function_from_graph_def(
    graph_def: tf.compat.v1.GraphDef,
    input_tensor_names: List[str],
    output_tensor_names: List[str],
) -> WrappedFunction:
    wrapped_import = tf.compat.v1.wrap_function(lambda: tf.compat.v1.import_graph_def(graph_def, name=""), [])
    imported_graph = wrapped_import.graph

    return wrapped_import.prune(
        tf.nest.map_structure(imported_graph.as_graph_element, input_tensor_names),
        tf.nest.map_structure(imported_graph.as_graph_element, output_tensor_names),
    )


class FrozenGraphLayer(tf.keras.layers.Layer):
    """
    Loads a TensorFlow V1 frozen graph (.pb) file from disk as a layer that
    can be used in another model.
    """

    def __init__(
        self,
        path_or_graph_def: Union[str, Dict[Any, Any], tf.compat.v1.GraphDef],
        input_tensor_names: List[str],
        output_tensor_names: List[str],
        name: Optional[str] = None,
    ):
        if isinstance(path_or_graph_def, tf.compat.v1.GraphDef):
            self.graph_def = path_or_graph_def
        elif isinstance(path_or_graph_def, dict):
            self.graph_def = tf.compat.v1.GraphDef()
            ParseDictToProtobuf(path_or_graph_def, self.graph_def)
        else:
            self.graph_def = tf.compat.v1.GraphDef()
            with tf.io.gfile.GFile(path_or_graph_def, "rb") as f:
                graph_str = f.read()
                self.graph_def.ParseFromString(graph_str)
            name = name or os.path.splitext(os.path.basename(path_or_graph_def))[0].replace(".", "_").replace("-", "_")

        self.input_tensor_names = input_tensor_names
        self.output_tensor_names = output_tensor_names
        self.concrete_function = _load_concrete_function_from_graph_def(
            self.graph_def,
            self.input_tensor_names,
            self.output_tensor_names,
        )

        super().__init__(name=name)

    def call(
        self,
        inputs: Union[tf.Tensor, List[tf.Tensor], Tuple[tf.Tensor]],
        training: Any = None,
    ) -> tf.Tensor:
        if isinstance(inputs, list) or isinstance(inputs, tuple):
            result = self.concrete_function(*inputs)
        else:
            result = self.concrete_function(inputs)

        # If we've only got one output tensor, return that tensor on its own as the layer's output.
        # This is the usual case. (Multi-output layers don't fit nicely into `Sequential` models, for instance.)
        if len(self.output_tensor_names) == 1:
            return result[0]
        return result

    def get_config(self) -> Dict[str, Any]:
        return {
            "path_or_graph_def": SerializeProtobufToDict(self.graph_def),
            "name": self.name,
            "input_tensor_names": self.input_tensor_names,
            "output_tensor_names": self.output_tensor_names,
        }


class SavedModelLayer(tf.keras.layers.Layer):
    """
    Loads a TensorFlow V2 SavedModel (saved_model.pb + variables/) from disk
    as a layer that can be used in another model. Use this instead of a Lambda
    layer to avoid variable capture issues and to produce a model that can be
    used correctly when re-saving a model.

    If multiple inputs (or outputs) exist, this layer will expect dictionaries
    on its input (or output):

        outputs = SavedModelLayer("./my_model/")({
            "some_input": <tensor>,
            "other_input": <tensor>
        })

        # Later:
        Dense(10)(outputs["some_output_name"])
    """

    def __init__(self, path_to_saved_model: str, name: Optional[str] = None):
        name = name or os.path.splitext(os.path.basename(path_to_saved_model))[0].replace(".", "_")
        super().__init__(name=name)
        self.model = tf.saved_model.load(path_to_saved_model)

    def call(self, _input: Union[tf.Tensor, Dict[str, tf.Tensor]]) -> Union[tf.Tensor, Dict[str, tf.Tensor]]:
        return self.model(_input)


def get_saved_model_input_tensors(saved_model_or_path: Union[tf.keras.Model, str]) -> List[tf.Tensor]:
    """
    Given a path to a SavedModel or an already loaded SavedModel,
    return a list of its input tensors. Useful for figuring out the
    names of the input tensors for use with SavedModelLayer.
    """
    if isinstance(saved_model_or_path, str):
        savedmodel = tf.saved_model.load(saved_model_or_path)
        model = savedmodel.signatures["serving_default"]
        model._backref = savedmodel  # Without this, the SavedModel will be GC'd too early
    else:
        model = saved_model_or_path
    if hasattr(model, "signatures"):
        model = model.signatures["serving_default"]

    return [tensor for tensor in model.inputs if tensor.dtype != "resource"]


def get_all_tensors_from_saved_model(saved_model_or_path: Union[tf.keras.Model, str]) -> List[tf.Tensor]:
    """
    Given a path to a SavedModel or an already loaded SavedModel,
    return a list of all of its tensors. Useful for figuring out the
    names of intermediate tensors for use with SavedModelLayer.

    To extract the output of a given Keras layer (which isn't stored in the
    SavedModel, as TensorFlow SavedModels don't acually save layer
    information), try something like:

    ```
        tensors = get_all_tensors_from_saved_model('./my-saved-model')

        layer_name = "some_named_layer"
        probable_output_of_layer = [t for t in tensors if layer_name in t.name][-1]

        # Should probably check that this is the model's input, as expected:
        probable_input_to_model = tensors[0]

        # Create a sub-graph of the loaded model from the input and output tensors you want:
        sub_graph = create_function_from_tensors(probable_input_to_model, probable_output_of_layer)

        # Use that sub-graph wherever you'd like:
        my_model = tf.keras.models.Sequential([
            tf.keras.layers.Lambda(lambda input_tensor: sub_graph(input_tensor)),
        ])

        # (Note that if you're extracting multiple inputs, your lambda function
        # must pass each input to sub_graph as a separate argument,
        # i.e.: `lambda input_tensors: sub_graph(*input_tensors)`.)
    ```

    """
    if isinstance(saved_model_or_path, str):
        savedmodel = tf.saved_model.load(saved_model_or_path)
        model = savedmodel.signatures["serving_default"]
        model._backref = savedmodel  # Without this, the SavedModel will be GC'd too early
    else:
        model = saved_model_or_path
    if hasattr(model, "signatures"):
        model = model.signatures["serving_default"]
    frozen_func, _ = convert_variables_to_constants_v2_as_graph(model)

    all_inputs_and_outputs: List[tf.Tensor] = sum(
        [list(op.inputs) + list(op.outputs) for op in frozen_func.graph.get_operations()],
        [],
    )

    # Using this to find unique tensors instead of set() because tf.Tensor is not hashable.
    seen_names = set()
    res = []
    for obj in all_inputs_and_outputs:
        if obj.name not in seen_names:
            seen_names.add(obj.name)
            res.append(obj)
    return res


def get_saved_model_output_tensors(saved_model_or_path: Union[tf.keras.Model, str]) -> List[tf.Tensor]:
    """
    Given a path to a SavedModel or an already loaded SavedModel,
    return a list of its output tensors. Useful for figuring out the
    names of the output tensors for use with SavedModelLayer.
    """
    if isinstance(saved_model_or_path, str):
        savedmodel = tf.saved_model.load(saved_model_or_path)
        model = savedmodel.signatures["serving_default"]
        model._backref = savedmodel  # Without this, the SavedModel will be GC'd too early
    else:
        model = saved_model_or_path
    if hasattr(model, "signatures"):
        model = model.signatures["serving_default"]
    return [tensor for tensor in model.outputs if tensor.dtype != "resource"]


def create_function_from_tensors(
    input_tensors: Union[tf.Tensor, List[tf.Tensor]],
    output_tensors: Union[tf.Tensor, List[tf.Tensor]],
) -> WrappedFunction:
    """
    Given two lists of tensors (input and output), this method will return a tf.function
    that can be used to feed inputs to the input tensors and take outputs from the output tensors.

    Example usage:

    ```
        tensors = get_all_tensors_from_saved_model('./my-saved-model')

        layer_name = "some_named_layer"
        probable_output_of_layer = [t for t in tensors if layer_name in t.name][-1]

        # Should probably check that this is the model's input, as expected:
        probable_input_to_model = tensors[0]

        # Create a sub-graph of the loaded model from the input and output tensors you want:
        sub_graph = create_function_from_tensors(probable_input_to_model, probable_output_of_layer)

        # Use that sub-graph wherever you'd like:
        my_model = tf.keras.models.Sequential([
            tf.keras.layers.Lambda(lambda input_tensor: sub_graph(input_tensor)),
        ])

        # (Note that if you're extracting multiple inputs, your lambda function
        # must pass each input to sub_graph as a separate argument,
        # i.e.: `lambda input_tensors: sub_graph(*input_tensors)`.)
    ```

    """
    if isinstance(input_tensors, tf.Tensor):
        input_tensors = [input_tensors]
    if isinstance(output_tensors, tf.Tensor):
        output_tensors = [output_tensors]

    if len(set([t.graph for t in (input_tensors + output_tensors)])) > 1:
        raise ValueError("All input and output tensors must be from the same graph.")

    graph = next(iter(input_tensors + output_tensors)).graph

    graph_input_names = [t.name for t in graph.inputs]

    try:
        return _load_concrete_function_from_graph_def(
            graph.as_graph_def(),
            [t.name for t in input_tensors],
            [t.name for t in output_tensors],
        )
    except UnliftableError as e:
        if len(graph_input_names) > len(input_tensors):
            raise ValueError(
                f"Could not create a function from input tensors {[t.name for t in input_tensors]} and output tensors"
                f" {[t.name for t in output_tensors]}. The graph that contains these tensors expects"
                f" {len(graph_input_names)} input(s), but only {len(input_tensors)} tensor(s) were passed - the outputs"
                " requested may depend on the missing inputs.\n\nUnderlying TensorFlow error:"
                f" {str(e).split(', e.g.:')[0]}."
            )
        else:
            raise


class TensorWrapperLayer(tf.keras.layers.Layer):
    """
    A really small layer class that allows passing input to an arbitrary input tensor
    and fetching input from an arbitrary output tensor. Useful for extracting parts of SavedModels.
    """

    def __init__(
        self,
        input_tensors: Union[tf.Tensor, List[tf.Tensor]],
        output_tensors: Union[tf.Tensor, List[tf.Tensor]],
        name: Optional[str] = None,
    ):
        super().__init__(name=name)
        self.func = create_function_from_tensors(input_tensors, output_tensors)

    def call(self, _input: Union[tf.Tensor, List[tf.Tensor]]) -> Union[tf.Tensor, List[tf.Tensor]]:
        if isinstance(_input, dict):
            raise NotImplementedError("Input to TensorWrapperLayer must be a single tensor or list of tensors.")
        elif isinstance(_input, list):
            return self.func(*_input)
        else:
            return self.func(_input)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2022 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import Any, Optional

import tensorflow as tf


class TrainingSpeedCallback(tf.keras.callbacks.Callback):
    """
    A tiny Keras callback to log to TensorBoard:
     - Seconds per [Epoch, Batch, Example]
     - Batches per Epoch
     - Examples per [Batch, Second]
    all grouped under the "Training Speed" header.
    """

    def __init__(
        self,
        tensorboard_writer: tf.summary.SummaryWriter,
        batches_per_epoch: int,
        examples_per_batch: Optional[int] = None,
    ):
        self.tensorboard_writer = tensorboard_writer
        self.batches_per_epoch = batches_per_epoch
        self.examples_per_batch = examples_per_batch

    def on_epoch_begin(self, epoch: int, logs: Any = None) -> None:
        self.epoch_begin_time = time.time()

    def on_epoch_end(self, epoch: int, logs: Any = None) -> None:
        epoch_end_time = time.time()
        seconds_per_epoch = epoch_end_time - self.epoch_begin_time
        seconds_per_batch = seconds_per_epoch / self.batches_per_epoch
        with self.tensorboard_writer.as_default():
            tf.summary.scalar(
                "Training Speed/Seconds Per Epoch",
                seconds_per_epoch,
                step=epoch,
            )
            tf.summary.scalar("Training Speed/Batches Per Epoch", self.batches_per_epoch, step=epoch)
            tf.summary.scalar(
                "Training Speed/Seconds Per Batch",
                seconds_per_batch,
                step=epoch,
            )
            if self.examples_per_batch:
                tf.summary.scalar(
                    "Training Speed/Seconds Per Example",
                    seconds_per_batch / self.examples_per_batch,
                    step=epoch,
                )
                tf.summary.scalar(
                    "Training Speed/Examples Per Second",
                    self.examples_per_batch / seconds_per_batch,
                    step=epoch,
                )
                tf.summary.scalar(
                    "Training Speed/Examples Per Batch",
                    self.examples_per_batch,
                    step=epoch,
                )
                tf.summary.scalar(
                    "Training Speed/Examples Per Epoch",
                    self.examples_per_batch * self.batches_per_epoch,
                    step=epoch,
                )



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2022 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License

import os
import tempfile

import numpy as np
import tensorflow as tf

from realbook.callbacks.speed import TrainingSpeedCallback
from realbook.layers.signal import Spectrogram

TEST_AUDIO = np.linspace(0, 1, num=22050 * 10)


def test_speed_callback() -> None:
    fake_data = tf.data.Dataset.zip(
        (
            tf.data.Dataset.from_tensor_slices([TEST_AUDIO]),
            tf.data.Dataset.from_tensor_slices([1]),
        )
    ).batch(1)

    model = tf.keras.Sequential(
        [
            tf.keras.Input(shape=(None,)),
            Spectrogram(),
            # Add a final "channel" dimension after the spectrogram
            tf.keras.layers.Lambda(lambda x: tf.expand_dims(x, -1)),
            tf.keras.layers.Dense(1),
        ]
    )
    model.compile(loss="binary_crossentropy")

    with tempfile.TemporaryDirectory() as tmpdir:
        writer = tf.summary.create_file_writer(tmpdir)
        speed_callback = TrainingSpeedCallback(writer, batches_per_epoch=1, examples_per_batch=1)

        model.fit(fake_data, callbacks=[speed_callback])
        writer.flush()

        all_points = 0
        for e in tf.compat.v1.train.summary_iterator(os.path.join(tmpdir, os.listdir(tmpdir)[0])):
            for v in e.summary.value:
                all_points += "Training Speed/" in v.tag
                value = tf.io.parse_tensor(v.tensor.SerializeToString(), tf.float32).numpy()
                if "Batches Per Epoch" in v.tag:
                    assert value == 1
                if "Examples Per Batch" in v.tag:
                    assert value == 1
        assert all_points == 7



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2022 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf


def log_base_b(x: tf.Tensor, base: int) -> tf.Tensor:
    """
    Compute log_b(x)
    Args:
        x : input
        base : log base. E.g. for log10 base=10
    Returns:
        log_base(x)
    """
    numerator = tf.math.log(x)
    denominator = tf.math.log(tf.constant(base, dtype=numerator.dtype))
    return numerator / denominator


class NormalizedLog(tf.keras.layers.Layer):
    """
    Takes an input with a shape of either (batch, x, y, z) or (batch, y, z)
    and rescales each (y, z) to dB, scaled 0 - 1.
    Only x=1 is supported.
    This layer adds 1e-10 to all values as a way to avoid NaN math.
    """

    def build(self, input_shape: tf.Tensor) -> None:
        self.squeeze_batch = lambda batch: batch
        rank = input_shape.rank
        if rank == 4:
            assert input_shape[1] == 1, "If the rank is 4, the second dimension must be length 1"
            self.squeeze_batch = lambda batch: tf.squeeze(batch, axis=1)
        else:
            assert rank == 3, f"Only ranks 3 and 4 are supported!. Received rank {rank} for {input_shape}."

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        inputs = self.squeeze_batch(inputs)  # type: ignore
        # convert magnitude to power
        power = tf.math.square(inputs)
        log_power = 10 * log_base_b(power + 1e-10, 10)

        log_power_min = tf.reshape(tf.math.reduce_min(log_power, axis=[1, 2]), [tf.shape(inputs)[0], 1, 1])
        log_power_offset = log_power - log_power_min
        log_power_offset_max = tf.reshape(
            tf.math.reduce_max(log_power_offset, axis=[1, 2]),
            [tf.shape(inputs)[0], 1, 1],
        )
        log_power_normalized = tf.math.divide_no_nan(log_power_offset, log_power_offset_max)

        return tf.reshape(log_power_normalized, tf.shape(inputs))



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2022 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License

import faulthandler
import time
from typing import Iterator, List, Tuple
import tensorflow as tf

import pytest
import pytest_mock
from realbook.callbacks.debugging import HangDebugCallback


def test_hang_debug_callback_does_not_fail() -> None:
    def generate_data() -> Iterator[Tuple[List[int], List[int]]]:
        for i in range(0, 10):
            yield ([1], [1])

    fake_data = tf.data.Dataset.from_generator(
        generate_data,
        output_signature=(
            tf.TensorSpec(shape=(1,), dtype=tf.int32),
            tf.TensorSpec(shape=(1,), dtype=tf.int32),
        ),
    )

    model = tf.keras.Sequential([tf.keras.Input(shape=(1)), tf.keras.layers.Dense(1)])
    model.compile(loss="binary_crossentropy")
    model.fit(fake_data, steps_per_epoch=1, epochs=10, callbacks=[HangDebugCallback()])


def test_hang_debug_callback_prints_stacks_if_slow(
    mocker: pytest_mock.MockerFixture, capsys: pytest.CaptureFixture
) -> None:
    mocker.patch("faulthandler.dump_traceback")
    print_stacks_time_multiple = 2
    abort_time_multiple = 100000
    hang_debug_callback = HangDebugCallback(print_stacks_time_multiple, abort_time_multiple)

    # Artificially speed up the callback's check frequency:
    hang_debug_callback.CHECK_INTERVAL_SECONDS = 0.1

    def generate_data() -> Iterator[Tuple[List[int], List[int]]]:
        duration = hang_debug_callback.CHECK_INTERVAL_SECONDS
        for i in range(0, 9):
            time.sleep(duration)
            yield ([1], [1])
        time.sleep(duration * print_stacks_time_multiple * 10)
        yield ([1], [1])

    fake_data = tf.data.Dataset.from_generator(
        generate_data,
        output_signature=(
            tf.TensorSpec(shape=(1,), dtype=tf.int32),
            tf.TensorSpec(shape=(1,), dtype=tf.int32),
        ),
    )

    model = tf.keras.Sequential([tf.keras.Input(shape=(1)), tf.keras.layers.Dense(1)])
    model.compile(loss="binary_crossentropy")
    model.fit(
        fake_data,
        steps_per_epoch=1,
        epochs=10,
        callbacks=[hang_debug_callback],
        verbose=0,
    )
    captured = capsys.readouterr()
    assert "Printing stacks" in captured.err
    faulthandler.dump_traceback.assert_called_once()  # type: ignore


def test_hang_debug_callback_aborts_if_really_slow(
    mocker: pytest_mock.MockerFixture, capsys: pytest.CaptureFixture
) -> None:
    mocker.patch("faulthandler.dump_traceback")
    print_stacks_time_multiple = 2
    abort_time_multiple = 3
    hang_debug_callback = HangDebugCallback(print_stacks_time_multiple, abort_time_multiple)

    # Artificially speed up the callback's check frequency:
    hang_debug_callback.CHECK_INTERVAL_SECONDS = 0.1

    # Patch the "abort" function to ensure we don't actually kill this Python process:
    hang_debug_callback.abort = mocker.MagicMock()  # type: ignore

    def generate_data() -> Iterator[Tuple[List[int], List[int]]]:
        duration = hang_debug_callback.CHECK_INTERVAL_SECONDS
        for i in range(0, 9):
            time.sleep(duration)
            yield ([1], [1])
        time.sleep(duration * abort_time_multiple * 10)
        yield ([1], [1])

    fake_data = tf.data.Dataset.from_generator(
        generate_data,
        output_signature=(
            tf.TensorSpec(shape=(1,), dtype=tf.int32),
            tf.TensorSpec(shape=(1,), dtype=tf.int32),
        ),
    )

    model = tf.keras.Sequential([tf.keras.Input(shape=(1)), tf.keras.layers.Dense(1)])
    model.compile(loss="binary_crossentropy")
    model.fit(
        fake_data,
        steps_per_epoch=1,
        epochs=10,
        callbacks=[hang_debug_callback],
        verbose=0,
    )
    captured = capsys.readouterr()
    assert "Aborting job" in captured.err
    hang_debug_callback.abort.assert_called()  # type: ignore



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2022 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Spotify"
__version__ = "1.0.0"
__email__ = "realbook@spotify.com"
__description__ = "Python libraries for easier machine learning on audio"
__url__ = "https://github.com/spotify/realbook"



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2022 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import threading
import time
import warnings
from typing import Any, Callable, Dict, cast

import tensorflow as tf

try:
    import psutil
    import nvsmi
except ImportError as e:
    raise ImportError(str(e) + " (did you 'pip3 install realbook[tensorboard]'?)")


class UtilizationCallback(tf.keras.callbacks.Callback):
    CHECK_FINISHED_TIMEOUT: float = 0.1
    RESOURCE_NAME: str = ""

    def __init__(self, tensorboard_writer: tf.summary.SummaryWriter, poll_period_seconds: float = 30):
        """
        A base class that can be used to create callbacks for measuring utilization of some resource.
        To override, override RESOURCE_NAME and resource_fn. resource_fn is a function that measures
        some resource and returns a percentage translating to amount utilized.

        This callback is not meant to be used in distributed training environments.

        Args:
            tensorboard_writer: A tensorboard SummaryWriter to write scalar summary events to.
            poll_period_seconds: The period (in seconds) at which to poll the resource.
        """
        if self.RESOURCE_NAME == "":
            raise ValueError("RESOURCE_NAME must be specified. Did you forget to overwrite it?")
        self.tensorboard_writer = tensorboard_writer
        # TF Can't really serialize objects for multiprocessing so...we get threads!
        self.measurement_thread = threading.Thread(
            target=self.measurement_poller,
            args=(
                self.tensorboard_writer,
                self.resource_fn,
                self.RESOURCE_NAME,
                poll_period_seconds,
            ),
            daemon=True,
        )
        self.finished = False

    def on_train_begin(self, logs: Dict[Any, Any]) -> None:
        """
        Runs at the beginning of training and starts the measurement thread.
        """
        self.measurement_thread.start()

    def on_train_end(self, logs: Dict[Any, Any]) -> None:
        """
        Runs at the end of training. Stops the measurement thread and flushes any buffered
        measurements.
        """
        self.finished = True
        self.measurement_thread.join()
        self.tensorboard_writer.flush()

    def measurement_poller(
        self,
        tensorboard_writer: tf.summary.SummaryWriter,
        resource_fn: Callable[[], float],
        resource_name: str,
        poll_period_seconds: float,
    ) -> None:
        """
        Function which is meant to run during the duration of an entire training run. This is
        what polls resource_fn(). This function takes arguments even though its run through
        threading in case we ever have the ability to switch it over to using multiprocessing.

        Args:
            tensorboard_writer: The SummaryWriter instance which holds tensorboard information.
            resource_fn: The function which runs the resource measurement logic.
            resource_name: The name of the resource to measure.
            poll_period_seconds: The period at which to poll in seconds.
        """

        reference_time = time.time()
        with tensorboard_writer.as_default():
            while not self.finished:
                last_update = time.time()
                util = resource_fn()
                tf.summary.scalar("Utilization/" + resource_name, util, int(last_update - reference_time))
                while not self.finished and (time.time() - last_update) < poll_period_seconds:
                    time.sleep(self.CHECK_FINISHED_TIMEOUT)

    @abc.abstractmethod
    def resource_fn(self) -> float:
        """
        A function to measure the utilization of a resource. Override this function with your
        resource measurement logic.

        Returns:
            The percentage used of a resource.
        """
        pass


class CpuUtilizationCallback(UtilizationCallback):
    """
    A utilization callback which measures cpu utilization.
    """

    RESOURCE_NAME = "CPU Utilization (%)"

    def resource_fn(self) -> float:
        return cast(float, psutil.cpu_percent())


class MemoryUtilizationCallback(UtilizationCallback):
    """
    A utilization callback which measures RAM utilization.
    """

    RESOURCE_NAME = "Memory Used (%)"

    def resource_fn(self) -> float:
        mem_stats = psutil.virtual_memory()
        # Don't use used as the psutil docs say that
        # "total - free does not necessarily match used."
        return (1.0 - float(mem_stats.available) / float(mem_stats.total)) * 100.0


class GpuResourceCallback(UtilizationCallback):
    """
    Base class for GPU utilization callbacks. This checks to see if a GPU is present.
    This class assumes that an NVIDIA GPU will be used.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        """
        Inititialize a GPU utilization callback. This callback will become a no-op
        if a GPU is not present of if nvidia-smi is not installed.
        """
        self.enable = True
        if not nvsmi.is_nvidia_smi_on_path():
            warnings.warn(
                "nvidia-smi is not on path. Please disable this callback. GPU usage data will be unavailable."
            )
            self.enable = False

        if self.enable and len(list(nvsmi.get_gpus())) == 0:
            warnings.warn("No GPUs detected. Please disable this callback. GPU usage data will be unavailable.")
            self.enable = False

        super().__init__(*args, **kwargs)

    def on_train_begin(self, logs: Dict[Any, Any]) -> None:
        """
        Runs at the beginning of training. Only starts the measurement thread if a GPU
        is available.
        """
        if self.enable:
            super().on_train_begin(logs)

    def on_train_end(self, logs: Dict[Any, Any]) -> None:
        """
        Runs at the end of training. Only stops the measurement thread if a GPU is
        available.
        """
        if self.enable:
            super().on_train_end(logs)


class GpuUtilizationCallback(GpuResourceCallback):
    """
    This class reports GPU utilization.
    """

    RESOURCE_NAME = "GPU Utilization (%)"

    def resource_fn(self) -> float:
        """
        Return: GPU utilization
        """
        return cast(float, next(nvsmi.get_gpus()).gpu_util)


class GpuMemoryUtilizationCallback(GpuResourceCallback):
    """
    This class reports GPU Memory utilization.
    """

    RESOURCE_NAME = "GPU Memory Utilization (%)"

    def resource_fn(self) -> float:
        """
        Return: GPU Memory utilization
        """
        return cast(float, next(nvsmi.get_gpus()).mem_util)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2022 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License

from typing import Any

import pytest
import numpy as np
import tensorflow as tf

from realbook.callbacks.spectrogram_visualization import SpectrogramVisualizationCallback
from realbook.layers.signal import Spectrogram


try:
    from contextlib import nullcontext
except ImportError:

    class nullcontext:  # type: ignore
        def __init__(self, obj: Any = None) -> None:
            self.obj = obj

        def __enter__(self) -> Any:
            return self.obj

        def __exit__(self, *args: Any, **kwargs: Any) -> None:
            pass


class FakeWriter:
    def as_default(self) -> Any:
        return nullcontext(self)

    def flush(self) -> None:
        pass


DEFAULT_SAMPLE_RATE = 22050
TEST_AUDIO = np.linspace(0, 1, num=DEFAULT_SAMPLE_RATE * 10)


def test_spectrogram_visualization_callback() -> None:
    fake_data = tf.data.Dataset.zip(
        (
            tf.data.Dataset.from_tensor_slices([TEST_AUDIO]),
            tf.data.Dataset.from_tensor_slices([1]),
        )
    ).batch(1)

    model = tf.keras.Sequential(
        [
            tf.keras.Input(shape=(None,)),
            Spectrogram(),
            tf.keras.layers.Dense(1),
        ]
    )
    model.compile(loss="binary_crossentropy")

    cb = SpectrogramVisualizationCallback(
        FakeWriter(),
        fake_data,
        sample_rate=DEFAULT_SAMPLE_RATE,
        raise_on_error=True,
    )

    model.fit(fake_data, callbacks=[cb])
    assert True


def test_callback_fails_on_unbatched_input() -> None:
    fake_data = tf.data.Dataset.zip(
        (
            tf.data.Dataset.from_tensor_slices([TEST_AUDIO]),
            tf.data.Dataset.from_tensor_slices([1]),
        )
    )

    model = tf.keras.Sequential(
        [
            tf.keras.Input(shape=(None,)),
            Spectrogram(),
            tf.keras.layers.Dense(1),
        ]
    )
    model.compile(loss="binary_crossentropy")

    # Cause the callback to error by passing in unbatched data.
    cb = SpectrogramVisualizationCallback(
        FakeWriter(),
        fake_data,
        sample_rate=DEFAULT_SAMPLE_RATE,
        raise_on_error=True,
    )

    with pytest.raises(AssertionError) as excinfo:
        model.fit(fake_data.batch(1), callbacks=[cb])
    assert "shape" in str(excinfo.value)


def test_callback_logs_but_doesnt_throw_by_default(caplog: pytest.LogCaptureFixture) -> None:
    fake_data = tf.data.Dataset.zip(
        (
            tf.data.Dataset.from_tensor_slices([TEST_AUDIO]),
            tf.data.Dataset.from_tensor_slices([1]),
        )
    )

    model = tf.keras.Sequential(
        [
            tf.keras.Input(shape=(None,)),
            Spectrogram(),
            tf.keras.layers.Dense(1),
        ]
    )
    model.compile(loss="binary_crossentropy")
    # Cause the callback to error by passing in unbatched data.
    cb = SpectrogramVisualizationCallback(FakeWriter(), fake_data, sample_rate=DEFAULT_SAMPLE_RATE)
    model.fit(fake_data.batch(1), callbacks=[cb])
    assert "AssertionError" in caplog.text
    assert "shape" in caplog.text


def test_fails_on_no_image_like_layers() -> None:
    fake_data = tf.data.Dataset.zip(
        (
            tf.data.Dataset.from_tensor_slices([TEST_AUDIO]),
            tf.data.Dataset.from_tensor_slices([1]),
        )
    ).batch(1)

    model = tf.keras.Sequential(
        [
            tf.keras.Input(shape=(None, 1)),
            tf.keras.layers.Dense(1),
        ]
    )
    model.compile(loss="binary_crossentropy")

    cb = SpectrogramVisualizationCallback(
        FakeWriter(),
        fake_data,
        sample_rate=DEFAULT_SAMPLE_RATE,
        raise_on_error=True,
    )

    with pytest.raises(ValueError) as excinfo:
        model.fit(fake_data, callbacks=[cb])
    assert isinstance(excinfo.value, ValueError)
    assert "spectrogram" in str(excinfo.value)


def test_flexible_with_input_shapes() -> None:
    fake_data = tf.data.Dataset.zip(
        (
            tf.data.Dataset.from_tensor_slices([TEST_AUDIO]),
            tf.data.Dataset.from_tensor_slices([1]),
        )
    ).batch(1)

    model = tf.keras.Sequential(
        [
            tf.keras.Input(shape=(None,)),
            Spectrogram(),
            # Add a final "channel" dimension after the spectrogram
            tf.keras.layers.Lambda(lambda x: tf.expand_dims(x, -1)),
            tf.keras.layers.Dense(1),
        ]
    )
    model.compile(loss="binary_crossentropy")

    cb = SpectrogramVisualizationCallback(
        FakeWriter(),
        fake_data,
        sample_rate=DEFAULT_SAMPLE_RATE,
        raise_on_error=True,
    )

    model.fit(fake_data, callbacks=[cb])
    assert True


def test_keras_functional_api_with_tfop_lambda() -> None:
    fake_data = tf.data.Dataset.zip(
        (
            tf.data.Dataset.from_tensor_slices([TEST_AUDIO]),
            tf.data.Dataset.from_tensor_slices([1]),
        )
    ).batch(1)

    _input = tf.keras.Input(shape=(None,))
    x = Spectrogram()(_input)
    # Add a final "channel" dimension after the spectrogram
    x = tf.expand_dims(x, -1)
    x = tf.keras.layers.Dense(1)(x)

    model = tf.keras.models.Model(inputs=_input, outputs=x)
    model.compile(loss="binary_crossentropy")

    cb = SpectrogramVisualizationCallback(
        FakeWriter(),
        fake_data,
        sample_rate=DEFAULT_SAMPLE_RATE,
        raise_on_error=True,
    )

    model.fit(fake_data, callbacks=[cb])
    assert True



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2022 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf
import torch
import numpy as np
import pytest
import librosa
import librosa.display

from typing import List, Tuple, Union

from realbook.layers import nnaudio as our_nnaudio
from nnAudio.Spectrogram import CQT2010v2

TEST_SAMPLE_RATE = 22050


# Test using this model directly, as well as wrapping it in a Lambda layer.
def get_parameterized_model_variants(
    match_torch_exactly_values: Tuple[bool, bool] = (True, False)
) -> List[tf.keras.layers.Layer]:
    possible_models = [
        our_nnaudio.CQT(match_torch_exactly=v, trainable=trainable)
        for v in match_torch_exactly_values
        for trainable in (True, False)
    ]

    return [
        item
        for models in [
            [tf.keras.Sequential([tf.keras.layers.InputLayer((TEST_SAMPLE_RATE,)), model])]
            + (
                [
                    tf.keras.Sequential(
                        [
                            tf.keras.layers.InputLayer((TEST_SAMPLE_RATE,)),
                            tf.keras.layers.Lambda(lambda x: model(x)),
                        ]
                    )
                ]
                # Using a layer with trainable weights inside a Lambda layer isn't supported.
                if not model.trainable
                else []
            )
            for model in possible_models
        ]
        for item in models
    ]


@pytest.mark.parametrize(
    "match_torch_exactly,threshold,trainable",
    (
        (True, 0.0001, False),
        (False, 4, False),
        (True, 0.0001, True),
        (False, 4, True),
    ),
)
def test_cqt(match_torch_exactly: bool, threshold: float, trainable: bool) -> None:
    # Ensure that the output of our CQT matches nnAudio's
    signal = librosa.chirp(fmin=32.70, fmax=22050, length=220500, linear=True)
    expected = np.transpose(CQT2010v2(verbose=False)(torch.tensor(signal, dtype=torch.float)).numpy(), (0, 2, 1))
    actual = our_nnaudio.CQT(match_torch_exactly=match_torch_exactly, trainable=trainable)(signal).numpy()

    assert expected.shape == actual.shape
    assert np.allclose(expected, actual, rtol=threshold, atol=threshold)


def build_layer(
    layer: tf.keras.layers.Layer, input_shape: Union[Tuple[int], Tuple[int, int]] = (1, TEST_SAMPLE_RATE)
) -> tf.keras.layers.Layer:
    layer.build(input_shape)
    return layer


def test_cqt_trainable_weights() -> None:
    assert not build_layer(our_nnaudio.CQT(trainable=False)).trainable
    assert not build_layer(our_nnaudio.CQT(trainable=False)).trainable_weights

    assert build_layer(our_nnaudio.CQT(trainable=True)).trainable
    assert len(build_layer(our_nnaudio.CQT(trainable=True)).trainable_variables) == 2
    assert len(build_layer(our_nnaudio.CQT(trainable=True)).trainable_weights) == 2


@pytest.mark.parametrize("train", (True, False))
def test_cqt_trainable_layers_change_on_training(train: bool) -> None:
    # Make a model that's trainable, then train it and ensure the weights change from the default.
    trainable_cqt = our_nnaudio.CQT(trainable=True)
    untrained_cqt = build_layer(our_nnaudio.CQT(trainable=True), (TEST_SAMPLE_RATE * 10,))

    model = tf.keras.Sequential(
        [
            tf.keras.layers.InputLayer((TEST_SAMPLE_RATE * 10,)),
            trainable_cqt,
            tf.keras.layers.Dense(1),
        ]
    )

    model.compile(loss="mse")
    if train:
        signal = librosa.chirp(32.70, TEST_SAMPLE_RATE, length=TEST_SAMPLE_RATE * 10, linear=True)
        noise = np.random.rand(*signal.shape).astype(signal.dtype)

        model.fit(np.array([signal, noise]), np.array([1, 0]), epochs=10, verbose=0)

    for untrained, trained in zip(untrained_cqt.trainable_weights, trainable_cqt.trainable_weights):
        weights_match_untrained = np.allclose(untrained.numpy(), trained.numpy())
        if train:
            assert not weights_match_untrained
        else:
            assert weights_match_untrained



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2022 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from typing import Any, Callable, Dict, Optional, Union

import librosa
import tensorflow as tf
import numpy as np


def _create_padded_window(
    window_fn: Callable[[int, tf.dtypes.DType], tf.Tensor],
    unpadded_window_length: int,
    fft_length: int,
) -> Callable[[int, tf.dtypes.DType], tf.Tensor]:
    lpad = (fft_length - unpadded_window_length) // 2
    rpad = fft_length - unpadded_window_length - lpad

    def padded_window(window_length: int, dtype: tf.dtypes.DType = tf.float32) -> tf.Tensor:
        # This is a trick to match librosa's way of handling window lengths < their fft_lengths
        # In that case the window is 0 padded such that the window is centered around 0s
        # In the Tensorflow case, the window is computed, multiplied against the frame and then
        # Right padded with 0's.
        return tf.pad(window_fn(unpadded_window_length, dtype=dtype), [[lpad, rpad]])  # type: ignore

    return padded_window


class Stft(tf.keras.layers.Layer):
    def __init__(
        self,
        fft_length: int = 2048,
        hop_length: Optional[int] = None,
        window_length: Optional[int] = None,
        window_fn: Callable[[int, tf.dtypes.DType], tf.Tensor] = tf.signal.hann_window,
        pad_end: bool = False,
        center: bool = True,
        pad_mode: str = "CONSTANT",
        name: Optional[str] = None,
        dtype: tf.dtypes.DType = tf.float32,
    ):
        """
        A Tensorflow Keras layer that calculates an STFT.
        The input is real-valued with shape (num_batches, num_samples).
        The output is complex-valued with shape (num_batches, time, fft_length // 2 + 1)

        Args:
            hop_length: The "stride" or number of samples to iterate before the start of the next frame.
            fft_length: FFT length.
            window_length: Window length. If None, then fft_length is used.
            window_fn: A callable that takes a window length and a dtype and returns a window.
            pad_end: Whether to pad the end of signals with zeros when the provided frame length and step produces
                a frame that lies partially past its end.
            center:
                If True, the signal y is padded so that frame D[:, t] is centered at y[t * hop_length].
                If False, then D[:, t] begins at y[t * hop_length].
            pad_mode: Padding to use if center is True. One of "CONSTANT", "REFLECT", or "SYMMETRIC" (case-insensitive).
            name: Name of the layer.
            dtype: Type used in calcuation.
        """
        super().__init__(trainable=False, name=name, dtype=dtype, dynamic=False)
        self.fft_length = fft_length
        self.window_length = window_length if window_length else self.fft_length
        self.hop_length = hop_length if hop_length else self.window_length // 4
        self.window_fn = window_fn
        self.final_window_fn = window_fn
        self.pad_end = pad_end
        self.center = center
        self.pad_mode = pad_mode

    def build(self, input_shape: tf.TensorShape) -> None:
        if input_shape.rank > 2:
            raise ValueError(
                "realbook.layers.signal.Stft received an input shape of "
                f"{input_shape}, but only supports inputs shaped "
                "like (num_samples,) or (num_batches, num_samples)."
            )

        if self.window_length < self.fft_length:
            self.final_window_fn = _create_padded_window(
                window_fn=self.window_fn,
                unpadded_window_length=self.window_length,
                fft_length=self.fft_length,
            )

        if self.center:
            self.spec = tf.keras.layers.Lambda(
                lambda x: tf.pad(
                    x,
                    [[0, 0] for _ in range(input_shape.rank - 1)] + [[self.fft_length // 2, self.fft_length // 2]],
                    mode=self.pad_mode,
                )
            )
        else:
            self.spec = tf.keras.layers.Lambda(lambda x: x)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        return tf.signal.stft(
            signals=self.spec(inputs),
            frame_length=self.fft_length,
            frame_step=self.hop_length,
            fft_length=self.fft_length,
            window_fn=self.final_window_fn,
            pad_end=self.pad_end,
        )

    def get_config(self) -> Dict[str, Any]:
        config: Dict[str, Any] = super().get_config().copy()
        config.update(
            {
                "fft_length": self.fft_length,
                "window_length": self.window_length,
                "hop_length": self.hop_length,
                "window_fn": self.window_fn,
                "pad_end": self.pad_end,
                "center": self.center,
                "pad_mode": self.pad_mode,
            }
        )
        return config


class Istft(tf.keras.layers.Layer):
    def __init__(
        self,
        fft_length: int = 2048,
        hop_length: Optional[int] = None,
        window_length: Optional[int] = None,
        window_fn: Callable[[int, tf.dtypes.DType], tf.Tensor] = tf.signal.hann_window,
        center: bool = True,
        name: Optional[str] = None,
        dtype: tf.dtypes.DType = tf.complex64,
    ):
        """
        A Tensorflow Keras layer that calculates an STFT.
        The input is complex-valued with shape (num_batches, num_frames, fft_length // 2 + 1)
            or (num_frames, fft_length // 2 + 1).
        The output is complex-valued with shape (num_batches, num_samples) or (num_samples).

        Args:
            hop_length: The "stride" or number of samples to iterate before the start of the next frame.
            fft_length: FFT length.
            window_length: Window length. If None, then fft_length is used.
                Warning. If this is less than fft_length then nans may appear in the output.
            window_fn: A callable that takes a window length and a dtype and returns a window.
                This will be converted to an inverse_window_fn.
                using tf.signal.inverse_stft_window_fn.
            center:
                If True, undo the centering done from stft (y(t) was padded so
                    that frame D[:, t] is centered at y[t * hop_length]).
                If False, don't modify the output of the istft.
                    You may need to trim the first and last hop_length samples
                    to match the original signal.
            name: Name of the layer.
            dtype: Type used in calcuation.
        """
        super().__init__(trainable=False, name=name, dtype=dtype, dynamic=False)

        # Because layer saves dtype as string???
        self.dtypes_type = dtype
        self.fft_length = fft_length
        self.window_length = window_length if window_length else self.fft_length
        self.window_fn = window_fn
        self.hop_length = hop_length if hop_length else self.window_length // 4
        self.center = center

        if self.fft_length <= self.hop_length:
            # If this is true then the inverse window function will contain a nan in some cases.
            raise ValueError("FFT Length must be less than or equal to hop length or else nans will appear.")

        if self.window_length < self.fft_length:
            warnings.warn(
                f"Istft layer received a window length ({self.window_length:,}) smaller than the provided FFT length"
                f" ({self.fft_length:,}).  NaN values may appear in this layer's output."
            )

        if not self.center:
            warnings.warn(
                "center=False has been passed to an Istft meaning perfect reconstruction is not obtainable."
                "The output in the range [hop_length:-hop_length] will still be valid."
            )

    def build(self, input_shape: tf.TensorShape) -> None:
        if input_shape.rank > 3 or input_shape.rank <= 1:
            raise ValueError(
                "realbook.layers.signal.Istft received an input shape of "
                f"{input_shape}, but only supports inputs shaped "
                "like (num_frames, num_bins) or (num_batches, num_frames, num_bins)."
            )

        self.window = _create_padded_window(self.window_fn, self.window_length, self.fft_length)(
            self.fft_length
        )  # type: ignore

        self.window_sum = librosa.filters.window_sumsquare(
            window=self.window.numpy(),
            n_frames=input_shape[0] if input_shape.rank == 2 else input_shape[1],
            win_length=self.window_length,
            n_fft=self.fft_length,
            hop_length=self.hop_length,
            dtype=self.dtypes_type.as_numpy_dtype,
        )

        self.window_sum = tf.constant(
            np.where(
                self.window_sum > np.finfo(self.dtypes_type.as_numpy_dtype).tiny,
                self.window_sum,
                np.ones_like(self.window_sum),
            ),
            self.dtypes_type.real_dtype,
        )

        self.slice_op = tf.keras.layers.Lambda(lambda x: x)
        if self.center:
            if input_shape.rank == 2:  # unbatched
                self.slice_op = tf.keras.layers.Lambda(
                    lambda x: x[int(self.fft_length // 2) : -int(self.fft_length // 2)]
                )
            else:  # batched
                self.slice_op = tf.keras.layers.Lambda(
                    lambda x: x[:, int(self.fft_length // 2) : -int(self.fft_length // 2)]
                )

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        istft = (
            tf.signal.overlap_and_add(self.window * tf.signal.irfft(inputs), frame_step=self.hop_length)
            / self.window_sum
        )
        return self.slice_op(istft)

    def get_config(self) -> Dict[str, Any]:
        config: Dict[str, Any] = super().get_config().copy()
        config.update(
            {
                "dtypes_type": self.dtypes_type,
                "fft_length": self.fft_length,
                "window_length": self.window_length,
                "hop_length": self.hop_length,
                "window_fn": self.window_fn,
                "center": self.center,
            }
        )
        return config


class Spectrogram(Stft):
    def __init__(
        self,
        power: int = 2,
        name: Optional[str] = None,
        dtype: tf.dtypes.DType = tf.float32,
        *args: Any,
        **kwargs: Any,
    ):
        """
        A Tensorflow Keras layer that calculates the magnitude spectrogram.
        The input is real-valued with shape (num_batches, num_samples).
        The output is real-valued with shape (num_batches, time, fft_length // 2 + 1)

        Args:
            power: Exponent to raise abs(stft) to.
            name: Name of the layer.
            dtype: Type used in calcuation.
            **kwargs: Any arguments that you'd pass to Stft
        """
        super().__init__(
            name=name,
            dtype=dtype,  # type: ignore
            *args,
            **kwargs,
        )
        self.power = power

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        return tf.math.pow(
            tf.math.abs(super().call(inputs)),
            self.power,
        )

    def get_config(self) -> Dict[str, Any]:
        config: Dict[str, Any] = super().get_config().copy()
        config.update(
            {
                "power": self.power,
            }
        )
        return config


class MelSpectrogram(Spectrogram):
    SLANEY = "slaney"  # slaney

    # the "type" of np.inf is np.float
    def __init__(
        self,
        sample_rate: int = 22050,
        fft_length: int = 2048,
        n_mels: int = 128,
        lower_edge_hertz: float = 0.0,
        upper_edge_hertz: Optional[float] = None,
        htk: bool = False,
        normalization: Optional[Union[float, int, str]] = SLANEY,
        *args: Any,
        **kwargs: Any,
    ):
        """
        A Tensorflow Keras layer that calculates the Mel Spectrogram of a signal.
        The input is real-valued with shape (num_batches, num_samples).
        The output is real-valued with shape (num_batches, time, n_mels)

        This function inherits from Spectrogram and supports all arguments from its
        __init__.

        Args:
            sample_rate: Sample rate of the input signal.
            fft_length: FFT length. If this is not a power of 2, the next power of 2 will be used as the FFT length.
            n_mels: How many bands in the resulting mel spectrum.
            lower_edge_hertz: Lower bound on the frequencies to be used in the Mel Spectrogram.
            upper_edge_hertz: Upper bound on the frequencies to be used in the Mel Spectrogram.
            htk: Use HTK formula for mel filters calculation instead of slaney.
            normalization: If 'slaney', divide the triangular mel weights by the width of the mel band
                (area normalization).
                Else, any value supported by [tf.norm](https://www.tensorflow.org/api_docs/python/tf/norm).
                If None, leave all the triangles aiming for a peak value of 1.0
        """
        fft_length = 1 << (fft_length - 1).bit_length()
        super().__init__(fft_length=fft_length, *args, **kwargs)
        self.n_mels = n_mels
        self.lower_edge_hertz = lower_edge_hertz
        self.upper_edge_hertz = upper_edge_hertz if upper_edge_hertz else float(sample_rate) / 2.0
        self.htk = htk
        self.sample_rate = sample_rate
        self.normalization = normalization

    def build(self, input_shape: tf.TensorShape) -> None:
        super().build(input_shape)

        self.mel_weight_matrix = tf.constant(
            librosa.filters.mel(
                sr=self.sample_rate,
                n_fft=self.fft_length,
                n_mels=self.n_mels,
                fmin=self.lower_edge_hertz,
                fmax=self.upper_edge_hertz,
                htk=self.htk,
                norm=self.normalization,
                dtype=np.float64,
            ).T,
            dtype=self.dtype,
        )

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        return tf.matmul(super().call(inputs), self.mel_weight_matrix)

    def get_config(self) -> Dict[str, Any]:
        config: Dict[str, Any] = super().get_config().copy()
        config.update(
            {
                "n_mels": self.n_mels,
                "lower_edge_hertz": self.lower_edge_hertz,
                "upper_edge_hertz": self.upper_edge_hertz,
                "htk": self.htk,
                "sample_rate": self.sample_rate,
                "normalization": self.normalization,
            }
        )
        return config



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2022 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

import librosa
import numpy as np
import pytest
import tensorflow as tf
from librosa.core.spectrum import _spectrogram
from librosa.feature.spectral import melspectrogram

from realbook.layers import signal


def test_stft_channels_should_raise() -> None:
    x = np.random.normal(0, 1, 1024)
    x = tf.expand_dims(tf.expand_dims(x, -1), 1)
    with pytest.raises(ValueError):
        signal.Stft(
            fft_length=256,
            hop_length=256,
            window_length=256,
            center=True,
        )(x)


@pytest.mark.parametrize(
    "center,input_length,fft_length,hop_length,win_length",
    [
        [True, 1024, 256, 256, 256],
        [False, 1024, 256, 256, 256],
        [False, 1024, 256, 256, 128],
        [False, 1024, 256, 512, 256],
    ],
)
def test_stft(center: bool, input_length: int, fft_length: int, hop_length: int, win_length: int) -> None:
    x = np.random.normal(0, 1, input_length)
    print(x.shape)
    librosa_stft = librosa.stft(
        y=x,
        n_fft=fft_length,
        hop_length=hop_length,
        win_length=win_length,
        center=center,
    ).T
    rgp_stft = signal.Stft(
        fft_length=fft_length,
        hop_length=hop_length,
        window_length=win_length,
        center=center,
    )(x).numpy()
    assert np.allclose(librosa_stft.real, rgp_stft.real, atol=1e-3, rtol=0)
    assert np.allclose(librosa_stft.imag, rgp_stft.imag, atol=1e-3, rtol=0)


def test_stft_batch() -> None:
    x = np.random.normal(0, 1, 1024)
    librosa_stft = librosa.stft(
        x,
        n_fft=256,
        hop_length=256,
        win_length=256,
        center=True,
    ).T
    rgp_stft = signal.Stft(
        fft_length=256,
        hop_length=256,
        window_length=256,
        center=True,
    )(tf.expand_dims(x, 0)).numpy()
    assert np.allclose(librosa_stft.real, rgp_stft.real, atol=1e-3, rtol=0)
    assert np.allclose(librosa_stft.imag, rgp_stft.imag, atol=1e-3, rtol=0)


def test_istft_channels_should_raise() -> None:
    x = np.random.normal(0, 1, 1024)
    x = tf.expand_dims(x, 0)
    x_stft = signal.Stft(
        fft_length=256,
        hop_length=256,
        window_length=256,
        center=True,
    )(x)
    with pytest.raises(ValueError):
        signal.Istft(
            fft_length=256,
            hop_length=256,
            window_length=256,
            center=True,
        )(tf.expand_dims(x_stft, -1))


@pytest.mark.parametrize(
    "center,input_length,fft_length,hop_length,win_length",
    [
        [True, 1024, 256, 128, 256],
        [True, 1024, 256, 64, 256],
        [False, 1024, 256, 128, 256],
        [False, 1024, 256, 64, 256],
        # TODO: The following case creates nans in the output. Fix that (nan => 0 does not count).
        # [False, 1024, 256, 128, 128],
        [False, 1024, 256, 256, 256],
        [True, 1025, 256, 128, 256],
        [False, 1025, 256, 128, 256],
        [True, 1023, 256, 128, 256],
        [False, 1023, 256, 128, 256],
    ],
)
def test_istft(center: bool, input_length: int, fft_length: int, hop_length: int, win_length: int) -> None:
    x = np.random.normal(0, 1, input_length)
    stft = signal.Stft(
        fft_length=fft_length,
        hop_length=hop_length,
        window_length=win_length,
        center=center,
    )(x)
    if fft_length == hop_length:
        with pytest.raises(ValueError):
            signal.Istft(
                fft_length=fft_length,
                hop_length=hop_length,
                window_length=win_length,
                center=center,
            )(stft).numpy()
    else:
        istft = signal.Istft(
            fft_length=fft_length,
            hop_length=hop_length,
            window_length=win_length,
            center=center,
        )(stft).numpy()

        librosa_istft = librosa.istft(
            stft.numpy().T,
            center=center,
            hop_length=hop_length,
            win_length=win_length,
        )

        assert np.allclose(
            istft,
            librosa_istft,
            atol=0.02,
            rtol=0,
        )

        if center:
            assert np.allclose(x[: len(istft)], istft, atol=1e-3, rtol=0)
        else:
            istft = istft[hop_length:-hop_length]
            x = x[hop_length:-hop_length]
            x = x[: len(istft)]
            assert np.allclose(x, istft, atol=1e-3, rtol=0)


def test_istft_batch() -> None:
    x = np.random.normal(0, 1, 1024)
    x_stft = signal.Stft(
        fft_length=256,
        hop_length=128,
        window_length=256,
        center=True,
    )(tf.expand_dims(x, 0))
    x_istft = signal.Istft(
        fft_length=256,
        hop_length=128,
        window_length=256,
        center=True,
    )(x_stft).numpy()
    assert np.allclose(x, np.squeeze(x_istft), atol=1e-3, rtol=0)


def test_spectrogram() -> None:
    x = np.random.normal(0, 1, 1024).astype(np.float32)
    librosa_spec, _ = _spectrogram(
        y=x,
        n_fft=1024,
        hop_length=256,
        power=2,
        win_length=256,
        center=True,
    )
    librosa_spec = librosa_spec.T
    rgp_spec = signal.Spectrogram(
        fft_length=1024,
        hop_length=256,
        window_length=256,
        center=True,
    )(x).numpy()
    assert np.allclose(librosa_spec, rgp_spec, atol=1e-2, rtol=0)


@pytest.mark.parametrize(
    "center,normalization,fmin,htk",
    [
        [True, None, 0.0, True],
        [True, None, 0.0, False],
        [True, None, 125.0, True],
        [False, signal.MelSpectrogram.SLANEY, 0.0, False],
        [False, 1, 0.0, False],
        [False, 2, 0.0, False],
        [False, np.inf, 0.0, False],
    ],
)
def test_mel_spectrogram(
    center: bool,
    normalization: Optional[Union[str, int, float]],
    fmin: float,
    htk: bool,
) -> None:
    x = np.random.normal(0, 1, 1024 * 8)
    librosa_spec = melspectrogram(
        y=x,
        n_fft=256,
        hop_length=128,
        power=2,
        win_length=256,
        center=center,
        htk=htk,
        norm=normalization,
        dtype=np.float32,
        fmin=fmin,
    ).T
    rgp_spec = signal.MelSpectrogram(
        fft_length=256, hop_length=128, center=center, normalization=normalization, htk=htk, lower_edge_hertz=fmin
    )(x).numpy()
    # These tolerances make look large, but the values we're comparing are in the 10e2-10e4 range.
    assert np.allclose(librosa_spec, rgp_spec, atol=1e-2, rtol=0)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2022 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile
from typing import Iterator, List, Tuple
import tensorflow as tf
from tensorflow.keras.layers import Dense, Concatenate
import numpy as np
from contextlib import contextmanager
from tensorflow.python.framework.convert_to_constants import (
    convert_variables_to_constants_v2_as_graph,
)
from tensorflow.lite.python.util import run_graph_optimizations, get_grappler_config

from realbook.layers.compatibility import (
    FrozenGraphLayer,
    SavedModelLayer,
    get_all_tensors_from_saved_model,
    TensorWrapperLayer,
)

NUM_INPUT_VALUES = 10


def keras_model_to_frozen_graph(
    keras_model: tf.keras.Model,
) -> Tuple[tf.compat.v1.GraphDef, List[tf.Tensor], tf.types.experimental.ConcreteFunction]:
    concrete_function = tf.function(keras_model).get_concrete_function(
        [tf.TensorSpec(i.shape, i.dtype) for i in keras_model.inputs]
    )
    frozen_func, graph_def = convert_variables_to_constants_v2_as_graph(concrete_function)
    input_tensors = [tensor for tensor in frozen_func.inputs if tensor.dtype != tf.resource]
    graph_def = run_graph_optimizations(
        graph_def,
        input_tensors,
        frozen_func.outputs,
        config=get_grappler_config([]),
        graph=frozen_func.graph,
    )
    return graph_def, input_tensors, frozen_func.outputs


@contextmanager
def keras_model_to_savedmodel(keras_model: tf.keras.Model) -> Iterator[str]:
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "temp_model")
        tf.saved_model.save(keras_model, path)
        yield path


def train_addition_model() -> tf.keras.models.Model:
    # Learn a model that sums numbers:
    model = tf.keras.Sequential(
        [
            tf.keras.layers.InputLayer(NUM_INPUT_VALUES),
            tf.keras.layers.Dense(1, name="my_layer"),
        ]
    )
    model.compile(loss="MSE")
    num_examples = 100000
    x = np.random.rand(num_examples, NUM_INPUT_VALUES)
    model.fit(x, np.sum(x, axis=-1))
    return model


def train_named_input_and_output() -> tf.keras.models.Model:
    # Learn a model that sums numbers:
    inputs = {"named_input": tf.keras.layers.Input(NUM_INPUT_VALUES)}
    outputs = {"named_output": tf.keras.layers.Dense(1)(inputs["named_input"])}
    model = tf.keras.models.Model(inputs=inputs, outputs=outputs)
    model.compile(loss="MSE")
    num_examples = 100000
    x = np.random.rand(num_examples, NUM_INPUT_VALUES)
    model.fit(x, np.sum(x, axis=-1))
    return model


def train_multi_input_multi_output() -> tf.keras.models.Model:
    """Train a dummy model to add numbers, but with multiple inputs and outputs: outAB = inA + inB, etc."""

    input_names = ["inA", "inB", "inC"]
    inputs = {input_name: tf.keras.layers.Input(1) for input_name in input_names}

    output_names = ["outAB", "outBC", "outAC"]
    concatenated = Concatenate()([v for _, v in sorted(inputs.items(), key=lambda t: t[0])])
    outputs = {
        "outAB": Dense(1)(concatenated),
        "outBC": Dense(1)(concatenated),
        "outAC": Dense(1)(concatenated),
    }

    model = tf.keras.models.Model(inputs=inputs, outputs=outputs)
    model.summary()

    model.compile(loss="MSE")
    num_examples = 100000
    inA, inB, inC = [np.random.rand(num_examples) for _ in range(3)]
    model.fit(
        {"inA": inA, "inB": inB, "inC": inC},
        {"outAB": inA + inB, "outBC": inB + inC, "outAC": inA + inC},
    )
    return model, input_names, output_names


def test_load_frozen_graph_as_layer() -> None:
    model = train_addition_model()
    x = np.random.rand(1, model.input_shape[-1])
    expected_output = model.predict(x)
    graph_def, input_tensors, output_tensors = keras_model_to_frozen_graph(model)

    reloaded_model = tf.keras.Sequential(
        [
            FrozenGraphLayer(
                graph_def,
                input_tensor_names=[t.name for t in input_tensors],
                output_tensor_names=[t.name for t in output_tensors],
            )
        ]
    )
    assert np.allclose(reloaded_model.predict(x), expected_output)


def test_load_frozen_graph_from_disk() -> None:
    model = train_addition_model()
    x = np.random.rand(1, model.input_shape[-1])
    expected_output = model.predict(x)

    graph_def, input_tensors, output_tensors = keras_model_to_frozen_graph(model)

    with tempfile.NamedTemporaryFile("wb") as f:
        f.write(graph_def.SerializeToString())
        f.flush()
        f.seek(0)

        reloaded_model = tf.keras.Sequential(
            [
                FrozenGraphLayer(
                    f.name,
                    input_tensor_names=[t.name for t in input_tensors],
                    output_tensor_names=[t.name for t in output_tensors],
                )
            ]
        )

    assert np.allclose(reloaded_model.predict(x), expected_output)


def test_reserialize_model() -> None:
    model = train_addition_model()
    x = np.random.rand(1, model.input_shape[-1])
    expected_output = model.predict(x)

    graph_def, input_tensors, output_tensors = keras_model_to_frozen_graph(model)

    with tempfile.NamedTemporaryFile("wb") as f:
        f.write(graph_def.SerializeToString())
        f.flush()
        f.seek(0)

        from_frozen = tf.keras.Sequential(
            [
                tf.keras.layers.InputLayer(NUM_INPUT_VALUES),
                FrozenGraphLayer(
                    f.name,
                    input_tensor_names=[t.name for t in input_tensors],
                    output_tensor_names=[t.name for t in output_tensors],
                ),
            ]
        )

    assert np.allclose(from_frozen.predict(x), expected_output)

    with tempfile.TemporaryDirectory() as tempdir:
        saved_path = os.path.join(tempdir, "output_model")
        from_frozen.save(saved_path)
        del from_frozen
        reloaded = tf.keras.models.load_model(saved_path)
    assert np.allclose(reloaded.predict(x), expected_output)


def test_load_multi_input_graph() -> None:
    # Learn a model that sums numbers, but uses multiple inputs:
    input_1 = tf.keras.layers.Input((NUM_INPUT_VALUES // 2,))
    input_2 = tf.keras.layers.Input((NUM_INPUT_VALUES // 2,))
    concat = tf.keras.layers.Concatenate()([input_1, input_2])
    output = tf.keras.layers.Dense(1)(concat)

    model = tf.keras.models.Model(inputs=[input_1, input_2], outputs=[output])
    model.compile(loss="MSE")
    num_examples = 100000
    x1 = np.random.rand(num_examples, NUM_INPUT_VALUES // 2)
    x2 = np.random.rand(num_examples, NUM_INPUT_VALUES // 2)
    model.fit((x1, x2), np.sum(np.concatenate((x1, x2), axis=-1), axis=-1))

    x1 = np.random.rand(1, NUM_INPUT_VALUES // 2)
    x2 = np.random.rand(1, NUM_INPUT_VALUES // 2)
    expected_output = model.predict([x1, x2])
    graph_def, input_tensors, output_tensors = keras_model_to_frozen_graph(model)

    reloaded_model = tf.keras.Sequential(
        [
            FrozenGraphLayer(
                graph_def,
                input_tensor_names=[t.name for t in input_tensors],
                output_tensor_names=[t.name for t in output_tensors],
            )
        ]
    )
    assert np.allclose(reloaded_model.predict([x1, x2]), expected_output)


def test_load_saved_model_as_layer_no_names() -> None:
    model = train_addition_model()
    x = np.random.rand(1, model.input_shape[-1])
    expected_output = model.predict(x)

    with keras_model_to_savedmodel(model) as saved_model_path:
        reloaded_model = tf.keras.Sequential([SavedModelLayer(saved_model_path)])
        assert np.allclose(reloaded_model.predict(x), expected_output)


def test_load_saved_model_as_layer_dict_input_and_output() -> None:
    model = train_named_input_and_output()
    x = np.random.rand(1, model.input_shape["named_input"][-1])
    expected_output = model.predict(x)

    with keras_model_to_savedmodel(model) as saved_model_path:
        layer = SavedModelLayer(saved_model_path)
    assert np.allclose(layer({"named_input": x})["named_output"], expected_output["named_output"])


def test_load_saved_model_as_layer_many_dict_input_and_output() -> None:
    model, input_names, output_names = train_multi_input_multi_output()
    x = {input_name: tf.constant([[i]], dtype=tf.float32) for i, input_name in enumerate(input_names)}
    expected_outputs = model.predict(x)

    with keras_model_to_savedmodel(model) as saved_model_path:
        layer = SavedModelLayer(saved_model_path)

    outputs = layer(x)

    assert all([key in outputs for key in output_names])
    for output_name in outputs.keys():
        assert np.allclose(
            outputs[output_name], expected_outputs[output_name]
        ), f"Output {output_name} did not match expected value after reloading."


def test_tensor_wrapper_layer() -> None:
    model = train_addition_model()
    x = np.random.rand(1, model.input_shape[-1])
    expected_output = model.predict(x)

    with keras_model_to_savedmodel(model) as saved_model_path:
        tensors = get_all_tensors_from_saved_model(saved_model_path)

        layer = TensorWrapperLayer(tensors[0].graph.inputs, [t for t in tensors if "my_layer" in t.name][-1])
        reloaded_model = tf.keras.Sequential([layer])
        assert np.allclose(reloaded_model.predict(x), expected_output)


def test_tensor_wrapper_layer_multiple_inputs() -> None:
    model, input_names, output_names = train_multi_input_multi_output()
    x = {input_name: tf.constant([[i]], dtype=tf.float32) for i, input_name in enumerate(input_names)}
    output_dict = model.predict(x)

    expected_outputs = [output_dict[output_name] for output_name in output_names]

    with keras_model_to_savedmodel(model) as saved_model_path:
        tensors = get_all_tensors_from_saved_model(saved_model_path)
        layer = TensorWrapperLayer(tensors[0].graph.inputs, tensors[0].graph.outputs)

    outputs = layer([value for _name, value in sorted(x.items(), key=lambda t: t[0])])  # type: ignore
    for expected, actual in zip(expected_outputs, outputs):
        assert np.allclose(actual, expected)
